#!/usr/bin/env python3

## usage
# at a level above emmer/
# python3 -m emmer.test.test_iteration

import unittest
from ..main.basic.math import NonDesityMatrix
from ..main.basic.read import RawDataImport
from ..main.advanced.iteration import MinusOneVNE, InfoRichCalling, reproducibility, reproducibility_summary, Kernal
from ..troubleshoot.warn.warning import *
from ..troubleshoot.err.error import ErrorCode4, ErrorCode6
import numpy
import pandas
import sys
import os


class TestMinusOneVNE(unittest.TestCase):

    def test_minusOneVNE(self, col_to_remove = 1, num_cpu = 1):
        print('\ntest_MinusOneVNE.minusOneVNE:')
        A = numpy.array([[1, 4, 5, 12], [5, 8, 9, 0], [6, 7, 11, 19]])
        InputA = MinusOneVNE(data = A, normalize = False, feature_names = ["col1", "col2", "col3", "col4"],  num_cpu = 1)
        my_result = InputA.minusOneVNE(col_to_remove = 1)
        #my_result[1] = numpy.round(my_result[1], decimals = 8)
        my_result[2] = numpy.round(my_result[2], decimals = 8)

        #expected_result = ['col2', 0.60207691]
        expected_result = [1, 'col2', 0.60207691]
        self.assertListEqual(my_result, expected_result)
        print('===========================================================')


    def test_minusOneResult(self):
        print('\ntest_MinusOneVNE.minusOneResult:')
        A = numpy.array([[1, 4, 5, 12], [5, 8, 9, 0], [6, 7, 11, 19]])
        InputA = MinusOneVNE(data = A, normalize = False, feature_names = ["col1", "col2", "col3", "col4"], num_cpu = 1)
        InputA.minusOneResult()
        my_result = list(numpy.round(numpy.array(InputA.result_summary['vNE']), decimals = 6))

        expected_result = [0.529410, 0.602077, 0.478102, 0.268291]
        self.assertListEqual(my_result, expected_result)
        print('===========================================================')


class TestInfoRichCalling(unittest.TestCase):

    def test_infoRichSelect(self):
        print('\ntest_InfoRichCalling.infoRichSelect:')
        print('        case 1: expect to work')
        file_name = 'emmer/data/test_case_1.csv'
        input_matrix = RawDataImport(file_name = file_name)
        input_matrix.readCSV()
        input_matrix.relativeAbundance()
        input_data = numpy.array(input_matrix.data)
        feature_names = list(input_matrix.feature_names)

        # filter == 'None'
        filtered_data = input_matrix

        info_rich_result = InfoRichCalling(data = filtered_data.data, current_feature_names = filtered_data.feature_names,
                                           upper_threshold_factor = 1, lower_threshold_factor = 1, normalize = False,
                                           num_cpu = 1, direct_from_result_summary = '')
        info_rich_result.infoRichSelect()
        my_result = list(info_rich_result.info_rich_feature['feature_name'])
        expected_result = ['col2', 'col3', 'col6']
        self.assertListEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('        case 2: raise error code 4')
        with self.assertRaises(ErrorCode4):
            file_name = 'emmer/data/test_case_1.csv'
            input_matrix = RawDataImport(file_name = file_name)
            input_matrix.readCSV()
            input_matrix.relativeAbundance()
            input_data = numpy.array(input_matrix.data)
            feature_names = list(input_matrix.feature_names)

            # filter == 'None'
            filtered_data = input_matrix

            info_rich_result = InfoRichCalling(data = filtered_data.data, current_feature_names = filtered_data.feature_names,
                                               upper_threshold_factor = 100, lower_threshold_factor = 100, normalize = False,
                                               num_cpu = 1, direct_from_result_summary = '', suppress = True)
            info_rich_result.infoRichSelect()
        print('===========================================================')


class Test_reproducibility_summary(unittest.TestCase):

    def test_reproducibility_summary(self):
        print('\ntest_reproducibility_summary:')
        file_name = 'emmer/data/test_case_1.csv'
        input_matrix = RawDataImport(file_name = file_name)
        input_matrix.readCSV()
        input_matrix.relativeAbundance()
        input_data = numpy.array(input_matrix.data)
        feature_names = list(input_matrix.feature_names)

        # filter == 'None'
        filtered_data = input_matrix

        infoRich_dict = {}
        info_rich_result = InfoRichCalling(data = filtered_data.data, current_feature_names = filtered_data.feature_names,
                                           upper_threshold_factor = 1, lower_threshold_factor = 1, num_cpu = 1,
                                           normalize = False, direct_from_result_summary = '')
        nrow = numpy.size(numpy.array(filtered_data.data), 0)

        info_rich_features_w_reproducibility = reproducibility_summary(filtered_matrix = filtered_data.data,
                                                                       infoRich_dict = reproducibility(InfoRichCalling_class = info_rich_result,
                                                                                                       infoRich_dict = infoRich_dict, nrow = nrow,
                                                                                                       basename = '', vNE_output_folder = '',
                                                                                                       output_file_tag = '', num_cpu = 1,
                                                                                                       normalize = False,
                                                                                                       direct_from_result_summary = ''))

        my_result = list(info_rich_features_w_reproducibility['occurrence'])
        expected_result = [4, 2, 3, 4, 1]
        self.assertListEqual(my_result, expected_result)

        my_result = list(info_rich_features_w_reproducibility['feature_name'])
        expected_result = ['col3', 'col5', 'col2', 'col6', 'col1']
        self.assertListEqual(my_result, expected_result)
        os.remove('__sub_0_detail_vNE.csv')
        os.remove('__sub_1_detail_vNE.csv')
        os.remove('__sub_2_detail_vNE.csv')
        os.remove('__sub_3_detail_vNE.csv')
        os.remove('__sub_4_detail_vNE.csv')
        os.remove('__sub_5_detail_vNE.csv')
        print('===========================================================')


class TestKernal(unittest.TestCase):

    def test_importAndProcess(self):
        ## use "HardFilter"
        print('        ---------------------------------------------------')
        print('        case 1: HardFilter (only set zero_tolerance)')
        print('             1.1: set detection limit correctly?')
        print('    file_names: test_case_3.csv')
        file_names = 'emmer/data/data_dir_1/test_case_1.csv'
        print('    detection_limit: 0')
        detection_limit = 0
        print('    tolerance: 0.6')
        tolerance = 0.6
        print('    filter: "HardFilter"')
        filter = 'HardFilter'
        print('    upper_lim: 1')
        upper_lim = 1
        print('    lower_lim: 1')
        lower_lim = 1
        print('    information-rich threshold: 1')
        infoRich_threshold = 1
        print('    quick_look_1: True')
        quick_look_1 = True
        print('    use_fractional_abundance: True')
        use_fractional_abundance = True
        print('    vNE_output_folder: ""')
        vNE_output_folder = ''
        print('    output_file_tag: out')
        output_file_tag = ''
        print('    normalize: False')
        normalize = False
        print('    num_cpu: 1')
        num_cpu = 1

        data = Kernal(file_name = file_names, detection_limit = detection_limit, tolerance = tolerance,
                      filter = filter, upper_lim = upper_lim, lower_lim = lower_lim, infoRich_threshold = infoRich_threshold,
                      quick_look = quick_look_1, use_fractional_abundance = use_fractional_abundance,
                      vNE_output_folder = vNE_output_folder, output_file_tag = output_file_tag,
                      num_cpu = num_cpu, notebook_name = '', normalize = normalize, neglect = True)
        data.importAndProcess()

        my_result = data.detection_limit
        expected_result = 0
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             1.2: set filter correctly?')
        my_result = data.filter
        expected_result = 'HardFilter'
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             1.3: remove correct features')
        my_result = list(data.filtered_data.data.columns.values)
        expected_result = ['col1', 'col2', 'col3', 'col5', 'col6']
        self.assertListEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             1.4: error handling')
        print('                  unexpected detection limit setting')
        file_names = 'emmer/data/data_dir_1/test_case_1.csv'
        print('    detection_limit: 100')
        detection_limit = 100
        print('    tolerance: 0.6')
        tolerance = 0.6
        print('    filter: "HardFilter"')
        filter = 'HardFilter'
        print('    upper_lim: 1')
        upper_lim = 1
        print('    lower_lim: 1')
        lower_lim = 1
        print('    information-rich threshold: 1')
        infoRich_threshold = 1
        print('    quick_look_1: True')
        quick_look_1 = True
        print('    use_fractional_abundance: True')
        use_fractional_abundance = True
        print('    vNE_output_folder: ""')
        vNE_output_folder = ''
        print('    output_file_tag: out')
        output_file_tag = ''
        print('    normalize: False')
        normalize = False
        print('    num_cpu: 1')
        num_cpu = 1

        with self.assertRaises(ErrorCode6):
            data = Kernal(file_name = file_names, detection_limit = detection_limit, tolerance = tolerance,
                          filter = filter, upper_lim = upper_lim, lower_lim = lower_lim, infoRich_threshold = infoRich_threshold,
                          quick_look = quick_look_1, use_fractional_abundance = use_fractional_abundance,
                          vNE_output_folder = vNE_output_folder, output_file_tag = output_file_tag,
                          num_cpu = num_cpu, neglect = True, normalize = normalize, notebook_name = '', suppress = True)
            data.importAndProcess()


        ## use "None"
        print('        ---------------------------------------------------')
        print('        case 2: None (only set zero_tolerance)')
        print('             2.1: set detection limit correctly?')
        print('    file_names: test_case_3.csv')
        file_names = 'emmer/data/data_dir_1/test_case_1.csv'
        print('    detection_limit: 0')
        detection_limit = 0
        print('    tolerance: 1')
        tolerance = 1
        print('    filter: "None"')
        filter = 'None'
        print('    upper_lim: 1')
        upper_lim = 1
        print('    lower_lim: 1')
        lower_lim = 1
        print('    information-rich threshold: 1')
        infoRich_threshold = 1
        print('    quick_look_1: True')
        quick_look_1 = True
        print('    use_fractional_abundance: True')
        use_fractional_abundance = True
        print('    vNE_output_folder: ""')
        vNE_output_folder = ''
        print('    output_file_tag: out')
        output_file_tag = ''
        print('    normalize: False')
        normalize = False
        if os.cpu_count() > 1:
            num_cpu = os.cpu_count() - 1

            print(f'    num_cpu: {num_cpu}')
        else:
            print('    num_cpu: 1')
            num_cpu = 1
            print('    your computer only have one CPU; unable to test multiprocessing function')

        data = Kernal(file_name = file_names, detection_limit = detection_limit, tolerance = tolerance,
                      filter = filter, upper_lim = upper_lim, lower_lim = lower_lim, infoRich_threshold = infoRich_threshold,
                      quick_look = quick_look_1, use_fractional_abundance = use_fractional_abundance,
                      vNE_output_folder = vNE_output_folder, output_file_tag = output_file_tag,
                      num_cpu = num_cpu, notebook_name = '', normalize = normalize, neglect = True)

        data.importAndProcess()

        my_result = data.detection_limit
        expected_result = 0
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             2.2: set filter correctly?')
        my_result = data.filter
        expected_result = 'None'
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             2.3: remove correct features')
        my_result = list(data.filtered_data.data.columns.values)
        expected_result = ['col1', 'col2', 'col3', 'col4', 'col5', 'col6']
        self.assertListEqual(my_result, expected_result)

        ## test detection limit
        print('        ---------------------------------------------------')
        print('        case 3: set detection limits')
        print('             3.1: set detection limits correctly')
        print('    file_names: test_case_3.csv')
        file_names = 'emmer/data/data_dir_1/test_case_1.csv'
        print('    detection_limit: 10')
        detection_limit = 10
        print('    tolerance: 1')
        tolerance = 1
        print('    filter: "None"')
        filter = 'None'
        print('    upper_lim: 1')
        upper_lim = 1
        print('    lower_lim: 1')
        lower_lim = 1
        print('    information-rich threshold: 1')
        infoRich_threshold = 1
        print('    quick_look_1: True')
        quick_look_1 = True
        print('    use_fractional_abundance: False')
        use_fractional_abundance = False
        print('    vNE_output_folder: ""')
        vNE_output_folder = ''
        print('    output_file_tag: out')
        output_file_tag = ''
        print('    normalize: False')
        normalize = False
        if os.cpu_count() > 1:
            num_cpu = os.cpu_count() - 1
            print(f'    num_cpu: {num_cpu}')
        else:
            print('    num_cpu: 1')
            num_cpu = 1
            print('    your computer only have one CPU; unable to test multiprocessing function')

        data = Kernal(file_name = file_names, detection_limit = detection_limit, tolerance = tolerance,
                      filter = filter, upper_lim = upper_lim, lower_lim = lower_lim, infoRich_threshold = infoRich_threshold,
                      quick_look = quick_look_1, use_fractional_abundance = use_fractional_abundance,
                      vNE_output_folder = vNE_output_folder, output_file_tag = output_file_tag,
                      num_cpu = num_cpu, notebook_name = '', normalize = normalize, neglect = True)

        data.importAndProcess()

        my_result = data.detection_limit
        expected_result = 10
        self.assertEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             3.2: set detection filter correctly')
        my_result = data.filter
        expected_result = 'None'
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             3.3: post-filter data')
        my_result = list(data.filtered_data.data['col2'])
        expected_result = [0, 0, 10, 0, 14, 14]
        self.assertListEqual(my_result, expected_result)
        print('===========================================================')


    def test_infoRichCallingAndReproducibility(self):
        print('\ntest_Kernal.infoRichCallingAndReproducibility:')
        print('        case 1: set quick_look (args.q) as False')
        print('                before consider the maximum number of information-rich feature')
        print('    file_names: test_case_1.csv')
        file_names = 'emmer/data/test_case_1.csv'
        print('    detection_limit: 0')
        detection_limit = 0
        print('    tolerance: 1')
        tolerance = 1
        print('    filter: "None"')
        filter = 'None'
        print('    upper_lim: 1')
        upper_lim = 0.5
        print('    lower_lim: 1')
        lower_lim = 0.5
        print('    information-rich threshold: 1')
        infoRich_threshold = 1
        print('    quick_look_2: False')
        quick_look_2 = False
        print('    use_fractional_abundance: True')
        use_fractional_abundance = True
        print('    vNE_output_folder: ""')
        vNE_output_folder = ''
        print('    output_file_tag: out')
        output_file_tag = ''
        print('    normalize: False')
        normalize = False
        print('    num_cpu: 1')
        num_cpu = 1

        data = Kernal(file_name = file_names, detection_limit = detection_limit, tolerance = tolerance,
                      filter = filter, upper_lim = upper_lim, lower_lim = lower_lim, infoRich_threshold = infoRich_threshold,
                      quick_look = quick_look_2, use_fractional_abundance = use_fractional_abundance,
                      vNE_output_folder = vNE_output_folder, output_file_tag = output_file_tag,
                      num_cpu = num_cpu, notebook_name = '', normalize = normalize, neglect = True)

        data.importAndProcess()
        data.infoRichCallingAndReproducibility()
        my_result = list(data.info_rich_features_w_reproducibility['feature_name'])
        expected_result = ['col1', 'col2', 'col3', 'col5', 'col4', 'col6']
        self.assertListEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('        case 2: maximum number of information-rich feature')
        print('             2.1 at quick look mode')
        print('                 ability to prevent report more information-rich feature than the maximum cap (Warning code 11)')
        print('    file_names: data/data_dir_4/group_A.csv')
        file_names = 'emmer/data/data_dir_4/group_A.csv'
        print('    detection_limit: 0.001')
        detection_limit = 0.001
        print('    tolerance: 0.33')
        tolerance = 0.33
        print('    filter: "HardFilter"')
        filter = 'HardFilter'
        print('    upper_lim: 2')
        upper_lim = 2
        print('    lower_lim: 0.2')
        lower_lim = 0.2
        print('    information-rich threshold: 2')
        infoRich_threshold = 2
        print('    quick_look_2: True')
        quick_look_2 = True
        print('    use_fractional_abundance: True')
        use_fractional_abundance = True
        print('    vNE_output_folder: ""')
        vNE_output_folder = ''
        print('    output_file_tag: out')
        output_file_tag = ''
        print('    normalize: False')
        normalize = False
        print('    num_cpu: 1')
        num_cpu = 1

        data = Kernal(file_name = file_names, detection_limit = detection_limit, tolerance = tolerance,
                      filter = filter, upper_lim = upper_lim, lower_lim = lower_lim, infoRich_threshold = infoRich_threshold,
                      quick_look = quick_look_2, use_fractional_abundance = use_fractional_abundance,
                      vNE_output_folder = vNE_output_folder, output_file_tag = output_file_tag,
                      num_cpu = num_cpu, notebook_name = '', normalize = normalize, neglect = True)
        data.importAndProcess()
        data.infoRichCallingAndReproducibility()

        my_result = data.list_of_info_rich_features
        expected_result = ['ASV_1', 'ASV_2', 'ASV_3', 'ASV_9', 'ASV_11', 'ASV_15', 'ASV_16', 'ASV_18', 'ASV_19', 'ASV_20', 'ASV_21', 'ASV_22', 'ASV_23']
        self.assertListEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             2.2 report WarningCode11')
        my_result = data.warning_code
        expected_result = '11'
        self.assertEqual(my_result, expected_result)


        ## Maximum number of information-rich feature (calculate reproducibility)
        ## also text the informaiton-rich feature calling
        print('        ---------------------------------------------------')
        print('             2.3 quick look mode off')
        print('             2.3.1 ability to prevent report more information-rich feature than the maximum cap (Warning code 11, 12)')
        print('    file_names: data/data_dir_4/group_A.csv')
        file_names = 'emmer/data/data_dir_4/group_A.csv'
        print('    detection_limit: 0.001')
        detection_limit = 0.001
        print('    tolerance: 0.33')
        tolerance = 0.33
        print('    filter: "HardFilter"')
        filter = 'HardFilter'
        print('    upper_lim: 0.2')
        upper_lim = 0.2
        print('    lower_lim: 0.2')
        lower_lim = 0.2
        print('    information-rich threshold: 2')
        infoRich_threshold = 2
        print('    quick_look_2: False')
        quick_look_2 = False
        print('    use_fractional_abundance: True')
        use_fractional_abundance = True
        print('    vNE_output_folder: ""')
        vNE_output_folder = ''
        print('    output_file_tag: out')
        output_file_tag = ''
        print('    normalize: False')
        normalize = False
        print('    num_cpu: 1')
        num_cpu = 1

        data = Kernal(file_name = file_names, detection_limit = detection_limit, tolerance = tolerance,
                      filter = filter, upper_lim = upper_lim, lower_lim = lower_lim, infoRich_threshold = infoRich_threshold,
                      quick_look = quick_look_2, use_fractional_abundance = use_fractional_abundance,
                      vNE_output_folder = vNE_output_folder, output_file_tag = output_file_tag,
                      num_cpu = num_cpu, notebook_name = '', normalize = normalize, neglect = True)
        data.importAndProcess()
        data.infoRichCallingAndReproducibility()

        my_result = list(data.info_rich_features_w_reproducibility['feature_name'])
        expected_result = ['ASV_1', 'ASV_2', 'ASV_3', 'ASV_11', 'ASV_22', 'ASV_23', 'ASV_9', 'ASV_18', 'ASV_19', 'ASV_21', 'ASV_15', 'ASV_16', 'ASV_27']
        self.assertListEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             2.3.2 report WarningCode12')
        my_result = data.warning_code
        expected_result = '12'
        self.assertEqual(my_result, expected_result)


        os.remove('test_case_1__sub_0_detail_vNE.csv')
        os.remove('test_case_1__sub_1_detail_vNE.csv')
        os.remove('test_case_1__sub_2_detail_vNE.csv')
        os.remove('test_case_1__sub_3_detail_vNE.csv')
        os.remove('test_case_1__sub_4_detail_vNE.csv')
        os.remove('test_case_1__sub_5_detail_vNE.csv')
        os.remove('group_A__detail_vNE.csv')
        os.remove('group_A__sub_0_detail_vNE.csv')
        os.remove('group_A__sub_1_detail_vNE.csv')
        os.remove('group_A__sub_2_detail_vNE.csv')
        os.remove('group_A__sub_3_detail_vNE.csv')
        os.remove('group_A__sub_4_detail_vNE.csv')
        os.remove('group_A__sub_5_detail_vNE.csv')
        os.remove('group_A__sub_6_detail_vNE.csv')
        os.remove('group_A__sub_7_detail_vNE.csv')
        os.remove('group_A__sub_8_detail_vNE.csv')
        os.remove('group_A__sub_9_detail_vNE.csv')
        os.remove('group_A__sub_10_detail_vNE.csv')
        os.remove('group_A__sub_11_detail_vNE.csv')
        os.remove('group_A__sub_12_detail_vNE.csv')
        print('===========================================================')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

## usage
# at a level above emmer/
# python3 -m emmer.test.test_input

from ..troubleshoot.err.error import ErrorCode7, ErrorCode9, ErrorCode10, ErrorCode28, ErrorCode31, ErrorCode33, ErrorCode36, ErrorCode37
from ..troubleshoot.inquire.input import EvaluateInput

from unittest.mock import patch
import unittest


class TestEvaluateInput(unittest.TestCase):

    def test_EvaluateInput(self):
        print('\ntest_EvaluateInput:')
        print('        case 1: raise error when the number of elements in input does not the match expecation')
        with self.assertRaises(ErrorCode10):
            confirmed_input = EvaluateInput(input = ['A', 'B', 'unexpected_additional_element'],
                                            set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                            suppress = True, expect = 0)
        print('===========================================================')


    def test_evaluateOption(self):
        print('\ntest_EvaluateInput.evaluateModeOption:')
        print('        case 1: missing expect argument')
        with self.assertRaises(ErrorCode7):
            confirmed_input = EvaluateInput(input = ['A', 'B'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                            suppress = True, expect = 0)
            confirmed_input.evaluateModeOption()


        print('        ---------------------------------------------------')
        print('        case 2: unexpected input')

        expected_dict = {'1': 'Individual',
                         '2': 'Permanova',
                         '3': 'RevisitThreshold',
                         '4': 'Reproducibility',
                         '5': 'Bifurication'}

        print('             2.1: unexpected option number')
        with self.assertRaises(ErrorCode9):
            confirmed_input = EvaluateInput(input = ['9999'], set = ['mode'],
                                            suppress = True, expect = expected_dict)
            confirmed_input.evaluateModeOption()
        #confirmed_input = EvaluateInput(input = ['9999'], set = ['mode'],
        #                                suppress = True, expect = expected_dict)
        #confirmed_input.evaluateModeOption()
        #my_result = processed_args.warning_code
        #expected_result = '9'
        #self.assertEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             2.2: unexpected option number')
        with self.assertRaises(ErrorCode9):
            confirmed_input = EvaluateInput(input = ['individual'], set = ['mode'],
                                            suppress = True, expect = expected_dict)
            confirmed_input.evaluateModeOption()
        #confirmed_input = EvaluateInput(input = ['individual'], set = ['mode'],
        #                                suppress = True, expect = expected_dict)
        #confirmed_input.evaluateModeOption()
        #my_result = processed_args.warning_code
        #expected_result = '9'
        #self.assertEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             3: expected to work')
        print('             3.1: output dictionary')
        confirmed_input = EvaluateInput(input = ['Individual'], set = ['mode'],
                                        suppress = True, expect = expected_dict)
        confirmed_input.evaluateModeOption()

        my_result = confirmed_input.map_input
        expected_result = {'mode': 'Individual'}
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             3.2: pass token')
        my_result = confirmed_input.passed
        expected_result = True
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


    def test_evaluateColor(self):
        print('\ntest_EvaluateInput.evaluateColor:')
        print('        case 1: EvaluateInput.expect should remain as default')
        with self.assertRaises(ErrorCode31):
            confirmed_input = EvaluateInput(input = ['red', 'blue'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                            suppress = True, expect = 'not_default')
            confirmed_input.evaluateColor()


        print('        ---------------------------------------------------')
        print('        case 2: not in mcolors.cnames.keys() or none or hex color')
        print('             2.1: not in mcolors.cnames.keys()')
        with self.assertRaises(ErrorCode28):
            confirmed_input = EvaluateInput(input = ['re', 'blue'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                            suppress = True, expect = 0)
            confirmed_input.evaluateColor()


        print('        ---------------------------------------------------')
        print('             2.1: not in hex color')
        with self.assertRaises(ErrorCode28):
            confirmed_input = EvaluateInput(input = ['red', '#ZZZZZZ'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                            suppress = True, expect = 0)
            confirmed_input.evaluateColor()


        print('        ---------------------------------------------------')
        print('        case 3: expect to work')
        print('             3.1: in mcolors.cnames.keys()')
        confirmed_input = EvaluateInput(input = ['red', 'blue'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                        suppress = True, expect = 0)
        confirmed_input.evaluateColor()

        my_result = confirmed_input.passed
        expected_result = True
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             3.1: in none')
        confirmed_input = EvaluateInput(input = ['none', 'none'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                        suppress = True, expect = 0)
        confirmed_input.evaluateColor()

        my_result = confirmed_input.passed
        expected_result = True
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             3.2: in hex color')
        confirmed_input = EvaluateInput(input = ['#ff0000', '#0000FF'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                        suppress = True, expect = 0)
        confirmed_input.evaluateColor()

        my_result = confirmed_input.passed
        expected_result = True
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


    def test_evaluateStatCluster(self):
        print('\ntest_EvaluateInput.evaluateStatCluster:')
        print('        case 1: EvaluateInput.expect should remain as default')
        with self.assertRaises(ErrorCode31):
            confirmed_input = EvaluateInput(input = ['A', 'B'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                            suppress = True, expect = 'not_default')
            confirmed_input.evaluateStatCluster()


        print('        ---------------------------------------------------')
        print('        case 2: only one cluster was given')
        with self.assertRaises(ErrorCode37):
            confirmed_input = EvaluateInput(input = ['A', 'A'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                            suppress = True, expect = 0)
            confirmed_input.evaluateStatCluster()


        print('        ---------------------------------------------------')
        print('        case 3: expect to work')
        confirmed_input = EvaluateInput(input = ['A', 'B'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                        suppress = True, expect = 0)
        confirmed_input.evaluateStatCluster()

        my_result = confirmed_input.passed
        expected_result = True
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


    def test_evaluateMarker(self):
        print('\ntest_EvaluateInput.evaluateMarker:')
        print('        case 1: EvaluateInput.expect should remain as default')
        with self.assertRaises(ErrorCode31):
            confirmed_input = EvaluateInput(input = ['A', 'B'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                            suppress = True, expect = 'not_default')
            confirmed_input.evaluateMarker()


        print('        ---------------------------------------------------')
        print('        case 2: not a marker')
        with self.assertRaises(ErrorCode33):
            confirmed_input = EvaluateInput(input = ['not_a_marker_A', 'not_a_marker_B'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                            suppress = True, expect = 0)
            confirmed_input.evaluateMarker()



        print('        ---------------------------------------------------')
        print('        case 3: expect to work')
        confirmed_input = EvaluateInput(input = ['o', 'H'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                        suppress = True, expect = 0)
        confirmed_input.evaluateMarker()

        my_result = confirmed_input.passed
        expected_result = True
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


    def test_evaluateGroupOption(self):
        print('\ntest_EvaluateInput.evaluateGroupOption:')
        print('        case 1: missing expect argument')
        with self.assertRaises(ErrorCode7):
            confirmed_input = EvaluateInput(input = ['A', 'B'], set = ['correpsonding_to_A', 'correpsonding_to_B'],
                                            suppress = True, expect = 0)
            confirmed_input.evaluateGroupOption()


        print('        ---------------------------------------------------')
        print('        case 2: unexpected input')

        expected_dict = {'1': 'Group',
                         '2': 'Individual'}

        print('             2.1: unexpected option number')
        with self.assertRaises(ErrorCode36):
            confirmed_input = EvaluateInput(input = ['9999'], set = ['color_by'],
                                            suppress = True, expect = expected_dict)
            confirmed_input.evaluateGroupOption()


        print('        ---------------------------------------------------')
        print('             2.2: unexpected option number')
        with self.assertRaises(ErrorCode36):
            confirmed_input = EvaluateInput(input = ['individual'], set = ['color_by'],
                                            suppress = True, expect = expected_dict)
            confirmed_input.evaluateGroupOption()


        print('        ---------------------------------------------------')
        print('             3: expected to work')
        print('             3.1: output dictionary')
        confirmed_input = EvaluateInput(input = ['Individual'], set = ['color_by'],
                                        suppress = True, expect = expected_dict)
        confirmed_input.evaluateGroupOption()

        my_result = confirmed_input.map_input
        expected_result = {'color_by': 'Individual'}
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


# https://stackoverflow.com/questions/21046717/python-mocking-raw-input-in-unittests

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

from ..troubleshoot.err.error import Error, ErrorCode44
from ..troubleshoot.inquire.input import InputCode6
import datetime
import time
import os

"""
Help to made log and record important results in  human-readable style
"""

## TODO: cover emmer.bake, error/warning/input, output dir

def initNoteBook(current_wd, script_name, script_version, tag, explicit, neglect, suppress = False):
    """
    Initate notebook for record keeping. Notebook name is stamped with time.
    To prevent initating multiple notebook for multiple emmer runs that are
    initated at the sametime, take from args.o (emmer.harvest and emmer.bake)
    to tag file name.

    When args.w == True (emmer.harvest and emmer.bake), allow user to input
    some personal massage (like the perpose of this analysis)

    Argument:
        current_wd -- Type: str
                      Current working directory
        script_name -- Type: str
                       Either 'emmer.harvest' or 'emmer.bake'
        script_version -- Type: str
                          Version of the script
        tag -- Type: str
               corresponding to agrs.o from 'emmer.harvest' or 'emmer.bake'
        explicit -- Type: boolean
                    corresponding to agrs.w from 'emmer.harvest' or 'emmer.bake'
        neglect -- Type: boolean
                   Suppress notebook initation when running unittest
        suppress -- Type: boolean
                    Suppress sys.exit() when error arised

    Reture:
        notebook_name -- Type: str
                         Notebook name with full path
    """

    if neglect == False:
        notebook_dir = os.path.join(current_wd, 'emmer_notebook')

        start_time = datetime.datetime.now()

        name_tag = ['_' + tag + '_' if tag != '' else ''][0]
        notebook_name = os.path.join(notebook_dir, 'emmer_' + str(time.mktime(start_time.timetuple()) + start_time.microsecond / 1E6) + name_tag + '_log.txt')

        ## make output dir if it does not exist
        if not os.path.exists(notebook_dir):
            os.makedirs(notebook_dir)

            try:
                if os.path.isfile(notebook_name):
                    raise Error(code = '44')
            except Error as e:
                raise ErrorCode44(suppress = suppress) from e

        notebook = open(notebook_name, "w")
        notebook.write('============================================================================================\n')
        notebook.write('Script: ' + script_name + '\n')
        notebook.write('Version: ' + script_version + '\n')
        notebook.write('Initate time: ' + str(start_time) + '\n')

        if explicit == True:
            notebook.write('\n' + InputCode6(suppress = False).input_string + '\n')

        notebook.write('============================================================================================\n')
        notebook.close()
    else:
        notebook_name = ''

    print('notebook initiate successfully...')
    return(notebook_name)


class UpdateNoteBook:
    """
    Update exitisng notebook.

    Arguments:
        notebook_name -- Type: str
                         Notebook name with full path
        neglect -- Type: boolean
                   Suppress notebook initation when running unittest

    Attributes:
        notebook_name -- Type: _io.TextIOWrapper
        neglect -- Type: boolean
    """

    def __init__(self, notebook_name, neglect):
        self.neglect = neglect
        if self.neglect == False:
            self.notebook = open(notebook_name, 'a')


    def updateArgs(self, args):
        if self.neglect == False:
            self.notebook.write('User input arguments:\n')
            args_in_str = str(args).replace('Namespace(', '    ').replace(')', '\n').replace(', ','\n    ')
            self.notebook.write(args_in_str)
            self.notebook.write('\n    Please refer to the Tutorial if you wish to know more about those arugments and their\n')
            self.notebook.write('    default settings. To view the Tutorial, run:\n')
            self.notebook.write('    python3 -m emmer.harvest -g\n')
            self.notebook.write('    python3 -m emmer.bake -g\n')
            self.notebook.write('============================================================================================\n')
            self.notebook.close()
            print('arguments import successfully...')


    def updateFilterResult(self, num_sample, num_feature, num_feature_removed, basename):
        if self.neglect == False:
            self.notebook.write(basename + '\n')
            self.notebook.write('    number of samples: ' + str(num_sample) + '\n')
            self.notebook.write('    number of features: ' + str(num_feature + num_feature_removed) + '\n')
            self.notebook.write('    number of features removed by pre-emmer data filter: ' + str(num_feature_removed) + '\n')
            self.notebook.close()


    def updateEmmerResult(self, num_info_rich):
        if self.neglect == False:
            self.notebook.write('    number of information-rich feature(s): ' + str(num_info_rich) + '\n')
            self.notebook.close()


    def updateMergeResult(self, merge_what, num_feature, norm_eigen):
        if self.neglect == False:
            self.notebook.write(merge_what + ' result\n')
            self.notebook.write('    number of feature(s): ' + str(num_feature) + '\n')
            self.notebook.write('    PC1: ' + str((norm_eigen[0])*100) + ' %\n')
            self.notebook.write('    PC2: ' + str((norm_eigen[1])*100) + ' %\n')

            if len(norm_eigen) > 2:
                self.notebook.write('    PC3: ' + str((norm_eigen[2])*100) + ' %\n')

            self.notebook.close()


    def updateProcrustesResult(self, which_dataset, procrustes_score, no_procrustes_score):
        if self.neglect == False:
            if no_procrustes_score == False:
                self.notebook.write('Procrustes test between ' + which_dataset + ' and original coordinates: ' + str(procrustes_score) + '\n')
            else:
                self.notebook.write(which_dataset + 'contains 0 feature. Skip procrustes test.')

            self.notebook.close()


    def updatePermanovaResult(self, set_seed, set_cluster, test_result):
        if self.neglect == False:
            self.notebook.write('set seed as: ' + str(set_seed) + '\n')
            self.notebook.write('grouping:\n' + str(set_cluster).replace('Name: cluster, dtype: object', '', 1) + '\n')
            self.notebook.write('\n')
            self.notebook.write(str(test_result))

        self.notebook.close()


    def updateRunTime(self, run_time):
        if self.neglect == False:
            self.notebook.write('============================================================================================\n')
            self.notebook.write('run time (exclude data merging and plotting): ' + str(run_time))
            self.notebook.close()



#!/usr/bin/env python3

## usage
# at a level above emmer/
# python3 -m emmer.test.test_math

from ..main.basic.math import NonDesityMatrix
from ..posthoc.visual.viewer import Projection
from ..troubleshoot.err.error import ErrorCode41

import numpy.testing
import unittest
import pandas
import numpy


class TestNonDesityMatrix(unittest.TestCase):

    def test_NonDesityMatrix(self):
        print('\ntest_NonDesityMatrix:')
        print('        case 1: dataset only have one row')
        A = numpy.array([[1, 4, 5, 12]])

        with self.assertRaises(ErrorCode41):
            MatrixA = NonDesityMatrix(data = A, normalize = False, suppress = True)

    def test_normEigvals(self):
        print('\ntest_NonDesityMatrix.normEigvals:')
        print('        case 1: use covariance matrix (= centered -> SVD)')
        print('        case 1.1: detect error generated when coding the arithmetic steps')
        A = numpy.array([[1, 4, 5, 12], [5, 8, 9, 0], [6, 7, 11, 19]])
        MatrixA = NonDesityMatrix(A, normalize = False)
        my_result = list(numpy.sort(numpy.round(MatrixA.normEigvals(), decimals = 8)))
        expected_result = [0.        , 0.1744152, 0.8255848]  # can also be validated in R; prcomp (scale = F)
        numpy.testing.assert_almost_equal(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('        case 1.2: detect error generating when using alternative arithmetic procedure to\neigendecomposition of the covariance matrix of the input matrix')
        eigenval = numpy.sort(numpy.linalg.eigvals(pandas.DataFrame(A).cov()))  # largest at the right most
        eigenval = eigenval[1:4]
        eigenval[eigenval < 10**-8] = 0
        expected_result = list(numpy.sort(numpy.round(eigenval/sum(eigenval), decimals = 8)))
        numpy.testing.assert_almost_equal(my_result, expected_result)
        print('===========================================================')

        print('        case 2: use correlation matrix (= centered -> scale -> SVD)')
        print('        case 2.1: detect error generated when coding the arithmetic steps')
        A = numpy.array([[1, 4, 5, 12], [5, 8, 9, 0], [6, 7, 11, 19]])
        MatrixA = NonDesityMatrix(A, normalize = True)
        my_result = list(numpy.sort(numpy.round(MatrixA.normEigvals(), decimals = 8)))
        expected_result = [0.        , 0.29255695, 0.70744305]
        numpy.testing.assert_almost_equal(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('        case 2.2: detect error generating when using alternative arithmetic procedure to\neigendecomposition of the correlation matrix of the input matrix')
        A = numpy.array([[1, 4, 5, 12], [5, 8, 9, 0], [6, 7, 11, 19]])
        matrixForSVD = numpy.corrcoef(numpy.transpose(A))
        eigenval = numpy.linalg.eigvalsh(matrixForSVD)
        eigenval = eigenval[1:4]
        eigenval[eigenval < 10**-8] = 0
        expected_result = list(numpy.sort(numpy.round(eigenval/sum(eigenval), decimals = 8))) # can also be validated in R; prcomp (scale = T)
        numpy.testing.assert_almost_equal(my_result, expected_result)

    def test_vNE(self):
        print('test_NonDesityMatrix.vNE:')
        print('        case 1: calculate the von Neuman entropy correctly')
        A = numpy.array([[1, 4, 5, 12], [5, 8, 9, 0], [6, 7, 11, 19]])
        MatrixA = NonDesityMatrix(A, normalize = False)
        my_result = numpy.round(MatrixA.vNE(), decimals = 8)

        expected_result = 0.66770591
        numpy.testing.assert_almost_equal(my_result, expected_result)
        print('===========================================================')


    def test_numZeroEig(self):
        # the ability to count the number of zero in the list of normalized eigenvalues.
        # (one from the list of normEigvals; one from ncol - nrow)
        print('test_NonDesityMatrix.numZeroEig:')
        print('        case 1: the ability to count the number of zero in the list of normalized eigenvalues.\n(one from the list of normEigvals; one from ncol - nrow)')
        A = numpy.array([[1, 4, 5, 12], [5, 8, 9, 0], [6, 7, 11, 19]])
        MatrixA = NonDesityMatrix(A, normalize = False)
        my_result = MatrixA.numZeroEig()
        self.assertEqual(my_result, 2)
        print('===========================================================')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# Usage:
# go one level above emmer/
# python3 -m emmer.test.test_viewer

from ..main.basic.math import NonDesityMatrix
from ..posthoc.visual.viewer import Projection

from pandas.util.testing import assert_frame_equal
import numpy.testing
import unittest
import pandas
import numpy


class TestProjection(unittest.TestCase):

    def test_Projection(self):
        print('\ntest_Projection:')
        print('        Case 1: colmean')
        A = [[.1, .4, .5, .12], [.05, .08, .09, .00], [.6, .7, .11, .19]]
        A_df = pandas.DataFrame(A, columns = ['col1', 'col2', 'col3', 'col4'], index = ['sample1', 'sample2', 'sample3'])
        MatrixA_for_transformation = Projection(merged_dataframe = A_df, normalize = False)

        colmean = list(numpy.round(numpy.array(MatrixA_for_transformation.colmean), decimals = 6))
        my_result = colmean
        #scaler = list(numpy.round(numpy.array(MatrixA_for_transformation.scaler), decimals = 6))
        #my_result = scaler

        expected_result = [0.25, 0.393333, 0.233333, 0.103333]
        self.assertListEqual(my_result, expected_result)


        print('===========================================================')
        print('        Case 2: V')
        nrow = len(A_df.index.values)
        S = numpy.tile(colmean, (nrow, 1))
        #S = numpy.tile(scaler, (nrow, 1))
        A_mean_norm = numpy.array(A_df) - S
        # A_mean_norm
        # row: sample; col: feature

        projection = numpy.matmul(A_mean_norm, numpy.array(MatrixA_for_transformation.V))
        # projection
        # row: feature; col: principal components

        my_result = pandas.DataFrame(numpy.round(-projection[:, 0:2], decimals = 5))
        expected_result = pandas.DataFrame([[-0.12439, 0.28011],
                                            [-0.35851, -0.20217],
                                            [ 0.48290, -0.07794]])  ## can also be validated in R; prcomp(A, scale = F)$x  # note: flip sign
        assert_frame_equal(my_result, expected_result)


        print('===========================================================')
        print('        Case 3: matrixS_adj')
        MatrixA_for_transformation.cleanSpec()
        my_result = MatrixA_for_transformation.S_adj
        expected_result = numpy.array([[0.6141608, 0, 0, 0],
        	                           [0, 0.3541278, 0, 0],
        	                           [0, 0, 5.777891e-17, 0]])  ## can also be validated in R; svd(scale(A, scale = F))
        numpy.testing.assert_almost_equal(my_result, expected_result, decimal = 5)


        print('===========================================================')
        print('        Case 4: matrixU')
        print('        Case 4.1: matrixU itself')
        my_result = list(MatrixA_for_transformation.U[:, 0])

        expected_result = [0.20254078, 0.58373539, -0.78627618]
        numpy.testing.assert_almost_equal(my_result, expected_result)


        print('===========================================================')
        print('        Case 5: spec_cleaned_data')
        print('        Case 5.1: a special matrix that has three singular values')
        print('                  In this special example, we got three non-zero singular values')
        print('                  (including the 10^-17 one), the result should equal to the')
        print('                  matrix before conducting SVD.')
        my_result = MatrixA_for_transformation.spec_cleaned_data

        expected_result = MatrixA_for_transformation.matrixForSVD
        numpy.testing.assert_almost_equal(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('        Case 5.2: a matrix that has more than three three singular values')
        print('                  Expect to consider only the first three singular values')
        print('                  when generating spec_cleaned_data. But spec_cleaned_data')
        print('                  should still have the same dimension as input matrix.')
        print('        Case 5.2.1: sanity check: make sure matrix B has more than three singular values')
        B = [[.1, .4, .5, .12], [.05, .08, .09, .00], [.6, .7, .11, .19], [.7, .7, .5, .2], [.1, .05, .9, .6]]
        B_df = pandas.DataFrame(B, columns = ['col1', 'col2', 'col3', 'col4'], index = ['sample1', 'sample2', 'sample3', 'sample4', 'sample5'])
        MatrixB_for_transformation = Projection(merged_dataframe = B_df, normalize = False)
        my_result = numpy.size(MatrixB_for_transformation.S, 0)

        self.assertEqual(my_result > 3, True)

        print('        ---------------------------------------------------')
        print('        Case 5.2.2: only consider the three singular values')
        MatrixB_for_transformation.cleanSpec()
        my_result = len(MatrixB_for_transformation.S_adj[numpy.nonzero(MatrixB_for_transformation.S_adj)])

        expected_result = 3
        self.assertEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('        Case 5.2.3: number of columns in spec_cleaned_data')
        my_result = numpy.size(MatrixB_for_transformation.spec_cleaned_data, 1)

        expected_result = 4  # ncol in input matrix B
        self.assertEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('        Case 5.2.4: number of rows in spec_cleaned_data')
        my_result = numpy.size(MatrixB_for_transformation.spec_cleaned_data, 0)

        expected_result = 5  # nrow in input matrix B
        self.assertEqual(my_result, expected_result)


        print('===========================================================')
        print('        Case 6: when set normalize == True')
        C = [[.1, .4, .5, .12], [.05, .08, .09, .00], [.6, .7, .11, .1]]
        C_df = pandas.DataFrame(C, columns = ['col1', 'col2', 'col3', 'col4'], index = ['sample1', 'sample2', 'sample3'])
        MatrixC_for_transformation = Projection(merged_dataframe = C_df, normalize = True)
        my_result = numpy.array(MatrixC_for_transformation.projection_df)

        expected_result = numpy.array([[-0.3329589,-1.4108734, 3.816392e-16],
                                       [ 1.7064763, 0.4766556, 1.325329e-15],
                                       [-1.3735174, 0.9342178,-1.963707e-15]])  ## can also be validated in R; prcomp(A, scale = F)$x  # note: flip sign
        numpy.testing.assert_almost_equal(my_result, expected_result)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

from ...toolbox.technical import *
from ...troubleshoot.err.error import *

import pandas
import numpy
import glob
import sys
import os


class RawDataImport:
    """
    Purpose:
        Take in a csv file name. Import data and transform to an input for NonDesityMatrix.
        Expect input has row.name (sample ID) and column name (feature name). Because without
        these information, it would be very easy for users to lose track. I hope my code would
        encourage this good practice.

    Arguments:
        file_name -- Type: str
                     file name with full path
        for_merging_file -- Type: boolean
                            Whether EMMER should raise ErrorCode when having '__' in the file name.
                            Set at 'False' when handling initial input.
        detection_limit -- Type: float
                           for self.detectionLimit()
        zero_tolerance_level -- Type: float
                                for self.zero_tolerance_level()

    Attributes:
        file_name -- Type: str
                     file name with full path
        basename -- Type: str
                    basename
        suppress -- Type: boolean
                    Whether EMMER will be terminated after rasing ErrorCode. Set at 'True' when
                    running unittest
        second_chance -- Type: boolean
                         Whether the user will have a second chance to change the input that
                         cause error or warning (This is a place holder; will add this option
                         in future update)
        raw_date -- Type: pandas.core.frame.DataFrame
        data -- Type: pandas.core.frame.DataFrame
        sample_id -- Type: list
                     row names
        feature_names -- Type: list
                         column header
        detection_limit -- Type: float
        zero_tolerance_level -- Type: float
    """

    def __init__(self, file_name, for_merging_file = False, suppress = False, second_chance = False):  ## TODO: retire second_chance

        self.file_name = file_name
        self.basename = os.path.basename(file_name)
        self.for_merging_file = for_merging_file
        self.suppress = suppress
        self.second_chance = second_chance

        # raise error when file names contain '__'
        if self.for_merging_file == False:
            try:
                if '__' in self.basename:
                    raise Error(code = '38')
            except Error as e:
                raise ErrorCode38(suppress = self.suppress) from e


    def readCSV(self):
        """
        Import data from a CSV file.
        """
        try:
            self.raw_data = pandas.read_csv(self.file_name, index_col = 0, header = 0)
        except FileNotFoundError as e:
            raise ErrorCode1(suppress = self.suppress) from e

        ## TODO: self.raw_data should only have number (no NA)
            ## TODO: or convert NA to zero


        # delete any column that only contain zeros
        self.data = self.raw_data.loc[:, (self.raw_data != 0).any(axis=0)]
        self.data = self.data.loc[(self.data != 0).any(axis=1), ]

        # update sample_id and feature_names
        self.sample_id = [str(element) for element in list(self.data.index.values)] # prevent using number as row names
        self.feature_names = [str(element) for element in list(self.data.columns.values)] # prevent using number as colnames

        if self.for_merging_file == False:
            try:
                if any('__' in element for element in self.sample_id):
                    raise Error(code = '39')
            except Error as e:
                raise ErrorCode39(suppress = self.suppress) from e

        # file_name should ended with .csv
        try:
            if not self.basename.endswith('.csv'):
                raise Error(code = '1')
        except Error as e:
            raise ErrorCode1(suppress = self.suppress) from e

        return(self.data)

    ## TODO:
    ## categorial data

    def deleteEmptyColRow(self):
        """
        Delete any column and row that only contains zeros.
        """
        self.data = self.data.loc[:, (self.data != 0).any(axis=0)]
        self.feature_names = list(self.data.columns.values)

        # delete any row that only contain zeros
        self.data = self.data.loc[(self.data != 0).any(axis=1), ]
        self.sample_id = list(self.data.index.values)
        return(self.data)

    def relativeAbundance(self):
        """
        Normalize counts for each feature by the sum of all counts in sampleself.

        Please run relativeAbundance() before setZero() and hardFilter() if you
        wish to use setZero() and hardFilter().
        """
        self.data = self.data.div(self.data.sum(axis = 1), axis = 0)

    def detectionLimit(self, detection_limit):
        self.detection_limit = detection_limit
        self.data[self.data < self.detection_limit] = 0
        self.deleteEmptyColRow()

    def hardFilter(self, zero_tolerance_level):
        """
        Manually define a hard filter to clean data based on the fraction of zero
        elements in each column.
        """
        r, c = self.data.shape
        mask = (self.data > 0).apply(numpy.count_nonzero) > numpy.round(zero_tolerance_level * r)
	    # True: the fraction of non-zero element is more then zero_tolerance_level
	    # False: otherwise
	    # test for all column and save the True/False in a vector

        self.data = self.data.loc[:, mask]
        self.data = self.deleteEmptyColRow()
        self.feature_names = list(self.data.columns.values)

	    # only keep the column if True
        return(self.data)


class RetrospectDataImport:
    """
    Import coordinate or parameter (grouping/clustering) or parameter (color/shape) setting
    file when excuting retrospect module.

    Arguments:
        file_name -- Type: str
                     file name with full path
        type -- Type: str
                Determine what the input file for

                'coordinate': Coordinates in PCA space
                'reproducibility': Information-rich calling reproducibility summary
                'vNE': Detail von Neumann entropy calculation result
                'data_color_shape_and_fill': For plotting

        dimension -- Type: str

                     'n': Not specified in input argument. Dimensionality will be determined
                          automatically
                     '2D': two dimensions
                     '3D': three dimensions

    Attributes:
        file_name -- Type: str
        coordinate -- Type: pandas.core.frame.DataFrame
        group -- Type: list
        group_set --
        individaul -- Type: list
        individual_set --
        coordinate_select_np --
        coordinate_w_info --
        reproducibility -- Type: pandas.core.frame.DataFrame
        vNE_summary --
        vNE_mean --
        vNE_sd --
        dimension --

    """

    def __init__(self, file_name, type, dimension = 'n', suppress = False):
        self.file_name = file_name

        try:
            imported_data = pandas.read_csv(self.file_name, index_col = 0, header = 0)
        except FileNotFoundError as e:
            raise ErrorCode1(suppress = suppress)

        try:
            if imported_data.isnull().values.any():
                raise Error(code = '45')
        except Error as e:
            raise ErrorCode45(suppress = suppress)

        ##--1--## coordinate
        if type == 'coordinate':
            self.coordinate = imported_data
            self.group = list([element.split("__")[0] for element in self.coordinate.index.values.tolist()])
            self.individual = list([element.split("__")[1] for element in self.coordinate.index.values.tolist()])
            self.group_set = set(self.group)
            self.individual_set = set(self.individual)

            try:
                if dimension == 'n':
                    if len(self.coordinate.columns.values) >= 3:
                        self.dimension = '3D'
                        coordinate_select = self.coordinate.iloc[:, 0:3].copy()
                    elif len(self.coordinate.columns.values) == 2:
                        self.dimension = '2D'
                        coordinate_select = self.coordinate.iloc[:, 0:2].copy()
                    else:
                        raise Error(code = '11')
                else:
                    self.dimension = dimension          # dimension in ['2D', '3D']
            except Error as e:
                raise ErrorCode11(suppress = suppress) from e

            self.coordinate_select_np = numpy.array(coordinate_select)
            self.coordinate_w_info = coordinate_select
            self.coordinate_w_info['group'] = self.group
            self.coordinate_w_info['individual'] = self.individual

        ##--2--## reproducibility
        elif type == 'reproducibility':
            repro_from_single_file = [col for col in imported_data.columns if col in ['repreducibility (%)', 'feature_name']]

            # case 1: summarized reproducibilty output
            #         row: information-rich taxa; column: group
            if len(set(repro_from_single_file)) == 0:
                self.reproducibility = imported_data
            # case 2: reproducibilty output from a single input file
            #         row: information-rich taxa; column: ['feature_name', 'occurrence', 'repreducibility (%)']
            elif len(set(repro_from_single_file)) == 2:
                self.reproducibility = pandas.DataFrame(data = numpy.array(imported_data[['repreducibility (%)']]),
                                                        columns = [os.path.basename(self.file_name)],
                                                        index = flattern(imported_data[['feature_name']].values))

        ##--3--## vNE
        elif type == 'vNE':
            self.vNE_summary = imported_data
            self.vNE_mean = numpy.mean(numpy.array(self.vNE_summary['vNE']))
            self.vNE_sd = numpy.std(numpy.array(self.vNE_summary["vNE"]), ddof = 1)

        ##--4--## data_color_shape_and_fill
        elif type == 'data_color_shape_and_fill':
            PC = [col for col in imported_data.columns if col in ['PC1', 'PC2', 'PC3']]

            try:
                if PC == ['PC1', 'PC2', 'PC3']:
                    self.dimension = '3D'
                elif PC == ['PC1', 'PC2']:
                    self.dimension = '2D'
                else:
                    raise Error(code = '29')
            except Error as e:
                raise ErrorCode29(suppress = suppress) from e


            targeted_col = [col for col in imported_data.columns if col in ['group', 'individual', 'edge_color', 'fill_color', 'shape']]
            try:
                if len(set(targeted_col)) != 5:
                    raise Error(code = '30')
            except Error as e:
                raise ErrorCode30(suppress = suppress) from e

            self.coordinate_w_info = imported_data

        ##--5--## v_df
        elif type == 'v_df':
            self.v_df = imported_data


        ##--6--## precent_explained
        elif type == 'precent_explained':
            self.precent_explained = imported_data

            try:
                if sorted(list(self.precent_explained.columns.values)) != sorted(['PC', 'percent explain', 'PC w percent explain']):
                    raise Error(code = '9')
            except Error as e:
                raise ErrorCode9(suppress = suppress) from e


class GetFiles:
    """
    Import the file names for mulitple csv files that saved under the same directory.

    Arguments:
        input_dir -- Type: str
        suppress -- Type: boolean
                    Whether exist program after error arised. Set at True when running
                    unittest
        second_chance -- Type: boolean  # TODO
                         Whether give user a second chance to re-enter the argument

    Attribute:
        input_files -- Type: list
                       A list of csv file name saved under input_dir
    """

    def __init__(self, input_dir, suppress = False, second_chance = False):
        self.input_files = glob.glob(os.path.join(input_dir, '*.csv'))

        # raise error when the folder has not csv file
        try:
            if len(self.input_files) < 1:
                raise Error(code = '1')
        except Error as e:
            raise ErrorCode1(suppress = suppress) from e


class MergeTargetedFiles:
    """
    Based on a mapping file, selected the targeted files and merged them

    Targeted file could be the individaul sequencing result. For example the first column
    could be all the ASV found in the sample, and the second column is the corresponding
    counts for each ASV. Expect target file has no row name and no header.

    Mapping file need to be a csv file that only has header, no row names and two column.
    The first column names corresponseding to the target file name. The name of the header should
    be ["file_name", "sample_id"]. The column should be the sample_id corresponding to each of the
    targeted file. This sample_id will eventually be used in RawDataImport() as the .sample_id

    The merged file will be a standard emmer.harvest args.i input file. Each row is a sample.
    Each column is a feature (e.g. ASV). Each element of the matrix is a numeric input (e.g. counts)

    Arguments:
        mapping_file -- Type: str
                        Path and name of the mapping file
        target_file_dir -- Type: str
                           Path to targeted files
        suppress -- Type: boolean
                    Whether end program after raising error. Set suppress = True
                    for unittest
        separate -- Type: str
                    Column in the targeted file separated by ?

    Attributes:
        mapping_file
        suppress
        map -- Type: pandas.DataFrame
               import from mapping file
        merged_target_file -- Type: pandas.DataFrame
                              dataframe ready for emmer.harvest analysis

    """

    def __init__(self, mapping_file, target_file_dir, separate, suppress):
        self.mapping_file = mapping_file
        self.target_file_dir = target_file_dir
        self.separate = separate
        self.suppress = suppress


    def getMap(self):
        self.map = pandas.read_csv(self.mapping_file, index_col = None, header = 0)

        try:
            if list(self.map.columns.values)[0:2] == ["file_name", "sample_id"]:
                pass
            else:
                raise Error(code = '40')
        except Error as e:
            raise ErrorCode40(suppress = self.suppress) from e


    def getTargetFile(self, targeted_file, col_name):
        df = pandas.read_csv(targeted_file, header = None, sep = self.separate)
        df = df.rename(columns = {0: 'feature', 1: col_name})
        return(df)


    def mergeFile(self):
        self.getMap()

        target_file_list = list(self.map['file_name'])
        sample_id_list = list(self.map['sample_id'])
        target_file_list_w_path = [os.path.join(self.target_file_dir, element) for element in target_file_list]

        for f in range(len(target_file_list_w_path)):
            sub = self.getTargetFile(targeted_file = target_file_list_w_path[f], col_name = sample_id_list[f])
            sub.columns.name = target_file_list[f]
            print(sub)

            if f == 0:
                merged_file = sub
            else:
                merged_file = merged_file.merge(sub, how = 'outer', on = 'feature')

        merged_file = merged_file.fillna(0)
        merged_file = merged_file.set_index('feature')
        self.merged_target_file = merged_file



#!/usr/bin/env python3

from ...main.basic.read import RetrospectDataImport
from .viewer import RetrospectPlot, Projection
from ...troubleshoot.err.error import *
from ...troubleshoot.warn.warning import WarningCode13
from ...troubleshoot.inquire.input import *

from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
import pandas
import numpy
import sys
import os

"""
Support emmer.bake Individual mode

plot individual samples onto the PCA plot. Allow better flexibility in aesthetics setting
"""


class IndividualArgs:
    """
    Take common arguments for Individual modes when running bake modules

    Argument:
        args -- Type: argparse.Namespace
                Store the user input parameters from command line
        current_wd -- Type: str
        suppress -- Type: boolean
                    Should emmer end program after error arise. Set at False when
                    running unittest
        silence -- Type: boolean

    Attributes:
        pca_coordinate -- Type: RetrospectDataImport class object (pandas Dataframe)
                          Updated PCA coordinates with plot setting (color, shape, etc.)
        warning_code -- Type: str
                        for unittest
        precent_explained -- Type: pandas Dataframe
                             Precent explained information for annotating PCs

    """

    def __init__(self, args, current_wd, suppress, silence):
        # annotate PCs
        if args.a:
            self.precent_explained = RetrospectDataImport(file_name = os.path.join(current_wd, args.a), type = 'precent_explained').precent_explained
        else:
            self.precent_explained = ''

        # have existing aesthetics parameter
        if args.p:
            try:
                if args.i:
                    raise WarningCode13(silence = silence)
                else:
                    self.warning_code = ''
            except WarningCode13:
                self.warning_code = '13'

            self.pca_coordinate = RetrospectDataImport(file_name = os.path.join(current_wd, args.p), type = 'data_color_shape_and_fill') # TODO: FileNotFoundError

        # do not have existing parameter
        else:
            try:
                if args.i and args.i.endswith('.csv'):
                    print(f'Input: {args.i}\n')
                    self.pca_coordinate = RetrospectDataImport(file_name = os.path.join(current_wd, args.i), type = 'coordinate')  # TODO: FileNotFoundError
                elif not args.i:
                    raise Error(code = '8')
                elif not args.i.endswith('.csv'):
                    raise Error(code = '8')
            except Error as e:
                if e.code == '8':
                    raise ErrorCode8(suppress = suppress) from e

            # TODO: percent explained

            #input_5_edge = InputCode5(add_msg = 'edge color of each dot', suppress = suppress, second_chance = True)
            input_5_edge = InputCode5(suppress = suppress)
            if input_5_edge.decision == ['Group'] or input_5_edge.decision == ['1']:
                print(self.pca_coordinate.group_set)
                input_1_group_edge = InputCode1(set = self.pca_coordinate.group_set, suppress = suppress)
                self.pca_coordinate.coordinate_w_info['edge_color'] = self.pca_coordinate.coordinate_w_info['group'].map(input_1_group_edge.color_dict)
            else:
                print(self.pca_coordinate.individual_set)
                input_1_individual_edge = InputCode1(set = self.pca_coordinate.individual_set, suppress = suppress)
                self.pca_coordinate.coordinate_w_info['edge_color'] = self.pca_coordinate.coordinate_w_info['individual'].map(input_1_individual_edge.color_dict)

            #input_5_fill = InputCode5(add_msg = 'fill color of each dot', suppress = suppress, second_chance = True)
            input_5_fill = InputCode5(suppress = suppress)
            if input_5_edge.decision == ['Group'] or input_5_edge.decision == ['1']:
                print(self.pca_coordinate.group_set)
                input_1_group_fill = InputCode1(set = self.pca_coordinate.group_set, suppress = suppress)
                self.pca_coordinate.coordinate_w_info['fill_color'] = self.pca_coordinate.coordinate_w_info['group'].map(input_1_group_fill.color_dict)
            else:
                print(self.pca_coordinate.individual_set)
                input_1_individual_fill = InputCode1(set = self.pca_coordinate.individual_set, suppress = suppress)
                self.pca_coordinate.coordinate_w_info['edge_color'] = self.pca_coordinate.coordinate_w_info['individual'].map(input_1_individual_fill.color_dict)

            # FIXME: cannot take shape
            #input_4_group_shape = InputCode4(set = self.pca_coordinate.group_set, add_msg = ['group'], suppress = suppress, second_chance = True)
            input_4_group_shape = InputCode4(set = self.pca_coordinate.group_set, suppress = suppress)
            self.pca_coordinate.coordinate_w_info['shape'] = self.pca_coordinate.coordinate_w_info['group'].map(input_4_group_shape.shape_dict)


def plotIndividual(args, current_wd, retrospect_dir, output_file_tag, suppress, silence):
        # python3 -m emmer.bake -m 'Individual' -i emmer/data/bake_data_dir_6/filtered_infoRich__PCA_coordinates.csv

        ## take-in args
        individual_args = IndividualArgs(args = args, current_wd = current_wd, suppress = suppress, silence = silence)

        ## generate PCA plot
        output_plot_name = os.path.join(retrospect_dir, (output_file_tag + '_individual_in_PCA.pdf'))

        if individual_args.pca_coordinate.dimension == '3D':
            plot_individauls_base_on_group = RetrospectPlot(coordinate_and_setting_df = individual_args.pca_coordinate.coordinate_w_info,
                                                            dimension = '3D', level = 'group', output_file_name = output_plot_name,
                                                            PC_annotation = individual_args.precent_explained)
        else:
            plot_individauls_base_on_group = RetrospectPlot(coordinate_and_setting_df = individual_args.pca_coordinate.coordinate_w_info,
                                                            dimension = '2D', level = 'group', output_file_name = output_plot_name,
                                                            PC_annotation = individual_args.precent_explained)

        plot_individauls_base_on_group.plot()

        output_file_name = os.path.join(retrospect_dir, (output_file_tag + '_retrospect_individaul_coloring_parameter.csv'))
        individual_args.pca_coordinate.coordinate_w_info.to_csv(output_file_name)



#!/usr/bin/env python3

from ...toolbox.technical import *
from ...main.basic.read import RawDataImport, RetrospectDataImport, GetFiles
from ...posthoc.visual.viewer import Projection
from ...troubleshoot.err.error import *

import statsmodels.api as sm
import pandas
import numpy
import math

"""
Support emmer.bake Bifurication mode

identify information-rich features that help to differentiate different groups on the PCA space.
"""

class BifuricationArgs:
    """
    Take common arguments for Bifurication modes when running bake modules

    Objective: so we can test and use @

    Argument:
        args -- Type: argparse.Namespace
                Store the user input parameters from command line
        suppress -- Type: boolean
                    Should emmer end program after error arise. Set at False when
                    running unittest
        second_chance -- Type: boolean
        current_wd -- Type: str
                      current working directory

    Attributes:
        list_of_info_rich -- Type: list
        input_files -- Type: list
                       List of input file with full path
    """
    def __init__(self, args, current_wd, suppress):

        # handle args.p (e.g.: information_rich_features_summary.csv)
        try:
            if args.p:
                 print(f'Input -p: {args.p}')
                 repro = RetrospectDataImport(file_name = os.path.join(current_wd, args.p), type = 'reproducibility')
            else:
                 raise Error(code = '12')
        except Error as e:
            raise ErrorCode12(suppress = suppress) from e
        self.list_of_info_rich = list(repro.reproducibility.index)

        # handle args.i (e.g. output/filtered_data/)
        try:
            if args.i:
                self.input_files = GetFiles(os.path.join(current_wd, args.i)).input_files

                if len(self.input_files) < 2:
                    raise Error(code = '21')
            else:
                raise Error(code = '21')
        except Error as e:
            raise ErrorCode21(suppress = suppress)

        def getArgsN(self):
            if self.args.n:
                self.normalize = True
            else:
                self.normalize = False

def linearRegressionPVal(target, data, silence_intersect = True):
    """
    Purpose:
        Get p values for each correlations coefficients and intercept in the linear
        regression model.

    Arguments:
        target -- Type: pandas.DataFrame
                  One column. The number of row corresponding to the number of sample.
                  Expect integer or float for each elements.
        data -- Type: pandas.DataFrame
                The number of column corresponding to the number of information-rich
                features. The number of row corresponding to the number of samples.
                Expect integer or float for each elements.
        silence_intersect -- Type: boolean
                             Report the p value for intersect or not. Please note that
                             when set silence_intersect = False, this function will still
                             calculate the linear regression model that has an intersect.


    Reture:
        pvalues -- Type: pandas.DataFrame
                   Row names (pandas.DataFrame.index) are the name of the feature

    Technical:
        Note that it is difficult to get this pieces of information from sklearn.linear_model
        The alternative is to use statsmodels package to conduct linear regression because
        it is easier to get that information from statsmodels output.
    """
    # consider intersect in the regression model. same as LinearRegression(fit_intercept=True)
    data_ = sm.add_constant(numpy.array(data))
        # now the first column of data_ is the intercept

    model = sm.OLS(target, data_)
    result = model.fit()

    pvalues = pandas.DataFrame(result.pvalues, columns = ['pval'])
    pvalues.index = ['intercept'] + list(data.columns)

    if silence_intersect == True:
        pvalues = pvalues.drop(pvalues.index[0])

    return(pvalues)


class DifferentiatingFeatures:
    """
    Purpose:
        Determine features that contribute to the significant difference of groups in
        PCA space.

    Arguments:
        Projection_class -- Type: Projection class object

        From Projection_class, DifferentiateGroup class requires:
        Projection_class.spec_cleaned_data -- Type: numpy.narray
                                              row: samples
                                              column: features
                                                      Although not restricted to specific set of
                                                      features, but it will make more sense for
                                                      the purpose of this analysis if those are
                                                      information-rich features.
        Projection_class.feature_names -- Type: list
                                          The colnames of the input matrix for Projection class.
                                          Also the colnames of the Projection_class.spec_cleaned_data
        Projection_class.sample_id -- Type: list
                                      The row names of the input matrix for Projection class.
                                      Also the row names of the Projection_class.spec_cleaned_data
    Attributes:
        spec_cleaned_df -- Type: pandas.DataFrame
        group_set -- Type: list
                     File names
        group_no_dict -- Type: dictionary
                         Key: self.group_set; value: group number (type: interger)
        differentiating_feature -- Type: list
    """

    def __init__(self, Projection_class):  # at cluster levels
        self.spec_cleaned_df = pandas.DataFrame(data = Projection_class.spec_cleaned_data,
                                                index = Projection_class.sample_id,
                                                columns = Projection_class.feature_names)
        self.spec_cleaned_df['group'] = tuple([element.split("__")[0] for element in Projection_class.sample_id])
        self.group_set = list(set(self.spec_cleaned_df['group']))

    def atGroup(self):
        group_set_num = len(self.group_set)
        group_no = list(range(group_set_num))
        self.group_no_dict = dict(zip(self.group_set, group_no))
        self.spec_cleaned_df['group_no'] = self.spec_cleaned_df['group'].map(self.group_no_dict)

        group_info = self.spec_cleaned_df[['group', 'group_no']]
        data_only = self.spec_cleaned_df.drop(columns = ['group', 'group_no'])

        pval_df = linearRegressionPVal(group_info['group_no'], data_only, silence_intersect = True)

        sign_pval_df = pval_df[(pval_df['pval'] < 0.1) & (pval_df['pval'] > 0.05)]
        sign_df = pandas.DataFrame(data = [1] * len(sign_pval_df.index),
                                   index = sign_pval_df.index,
                                   columns = ['sign_level'])

        high_sign_pval_df = pval_df[pval_df['pval'] < 0.05]
        high_sign_df = pandas.DataFrame(data = [2] * len(high_sign_pval_df.index),
                                        index = high_sign_pval_df.index,
                                        columns = ['sign_level'])

        merged_sign_df = pandas.concat([high_sign_df, sign_df], axis = 0, sort = True)

        self.differentiating_feature = list(merged_sign_df.index)
        self.differentiating_feature_df = merged_sign_df


def identifiedFeatures(args, current_wd, retrospect_dir, output_file_tag, suppress):
        ## take-in args
        bifurication_args = BifuricationArgs(args = args, current_wd = current_wd, suppress = suppress)

        for f in range(len(bifurication_args.input_files)):
            input_matrix = RawDataImport(file_name = bifurication_args.input_files[f], for_merging_file = True, suppress = False, second_chance = False)
            input_matrix.readCSV()

            info_rich = [value for value in input_matrix.data.columns.values if value in bifurication_args.list_of_info_rich]
            input_data = input_matrix.data[info_rich]

            if f == 0:
                merged_dataframe = input_data
            else:
                merged_dataframe = pandas.concat([merged_dataframe, input_data], axis = 0, sort = True)

        merged_dataframe = merged_dataframe.fillna(0)

        ## find feature(s) that responsible for differentiating groups in PCA space
        projection = Projection(merged_dataframe = merged_dataframe, normalize = bifurication_args.normalize)
        projection.cleanSpec()

        differentiating_features = DifferentiatingFeatures(projection)
        differentiating_features.atGroup()

        if output_file_tag != '':
            tag = [output_file_tag]
        else:
            tag = ['sign_level']

        output_file_name = os.path.join(retrospect_dir, (output_file_tag + '_differentiating_feature.csv'))

        differentiating_features.differentiating_feature_df.columns = tag

        print('Features that drive the separation among groups on PC space are:')
        print('(level 2: p value < 0.05; level 1: 0.05 < p value < 0.1)')
        print(differentiating_features.differentiating_feature_df)

        print('Output file:')
        print(output_file_name)
        differentiating_features.differentiating_feature_df.to_csv(output_file_name)

        # TODO: export raw_p_val



#!/usr/bin/env python3


from ...main.basic.read import RawDataImport, RetrospectDataImport, GetFiles
from ...main.advanced.iteration import InfoRichCalling, reproducibility_summary
from ...toolbox.technical import flattern, emptyNumpyArray, toFloat, floatRange
from ...troubleshoot.inquire.input import *
from ...troubleshoot.err.error import *

from ..visual.projection import ProjectionArgs, projectNew
from ..visual.viewer import RetrospectPlot, Projection
from ..visual.individual import plotIndividual

from .bifurication import identifiedFeatures
from .reproducibility import reproSummary
from .permanova import permanovaResult

from multiprocessing import Pool
from scipy.spatial import procrustes
from sklearn import linear_model
from tqdm import tqdm

import itertools
import argparse
import pandas
import numpy
import sys
import os


"""
Support emmer.bake RevisitThreshold mode
"""


def evaluateInputTuple(input_tuple, suppress = False, second_chance = False):
    """
    Make sure args.t, args.u, args.l input are logical
    """
    input_parameters = tuple(map(toFloat, input_tuple.split(',')))

    ## tuple length should be 3
    try:
        if len(input_parameters) != 3:
            raise Error(code = '17')
    except Error as e:
        if e.code == '17':
            raise ErrorCode17(suppress = suppress) from e

    ## check the first element (max) and the second element (min) values
    # input max and min need to be greater than zero
    if input_parameters != (0,0,0):
        try:
            if input_parameters[0] <= 0 or input_parameters[1] <= 0:
                raise Error(code = '20')
        except Error as e:
            raise ErrorCode20(suppress = suppress) from e

        # max need to be greater or equal to min
        try:
            if input_parameters[0] > input_parameters[1]:
                # in this case, except the third element (increment) to be...
                if input_parameters[2] <= 0:
                    raise Error(code = '24')

                # lower than max
                if input_parameters[0] <= input_parameters[2]:
                    raise Error(code = '25')

                # (input_parameters[0] - input_parameters[1]) % input_parameters[2] = 0
                if numpy.equal((input_parameters[0] - input_parameters[1]) % input_parameters[2], 0) == False:
                    raise Error(code = '22')

            elif input_parameters[0] == input_parameters[1]:
                # in this case, except the third element (increment) to be 0
                if input_parameters[2] != 0:
                    raise Error(code = '26')
            else:
                raise Error(code = '19')

        except Error as e:
            if e.code == '19':
                raise ErrorCode19(suppress = suppress) from e
            elif e.code == '22':
                raise ErrorCode22(suppress = suppress) from e
            elif e.code == '24':
                raise ErrorCode24(suppress = suppress) from e
            elif e.code == '25':
                raise ErrorCode25(suppress = suppress) from e
            elif e.code == '26':
                raise ErrorCode26(suppress = suppress) from e

    return(input_parameters)


class RevisitThresholdArgs:
    """
    Take common arguments for RevisitThreshold modes when running bake modules

    Objective: so we can test and use @

    Argument:
        args -- Type: argparse.Namespace
                Store the user input parameters from command line
        suppress -- Type: boolean
                    Should emmer end program after error arise. Set at False when
                    running unittest
        silence -- Type: boolean
        second_chance -- Type: boolean

    Attributes:
        args -- Type: argparse.Namespace
                Store the user input parameters from command line
        suppress -- Type: boolean
        silence -- Type: boolean
        second_chance -- Type: boolean
        detail_vNE --
        data_file --
        tuple_u --
        tuple_l --
        tuple_t --
        normalize -- Type: boolean
                     Scale each column in the mean centered data based on its standard deviation
                     before SVD

    """
    def __init__(self, args, current_wd, suppress, silence):
        self.args = args
        self.current_wd = current_wd
        self.suppress = suppress
        self.silence = silence


    #def getArgsN(self):
    def getArgsE(self):
        try:
#            if self.args.n:
            if self.args.e:
#                input_dir = os.path.join(self.current_wd, self.args.n)
                input_dir = os.path.join(self.current_wd, self.args.e)
                self.detail_vNE = GetFiles(input_dir = input_dir)
            else:
                raise Error(code = '13')
        except Error as e:
            raise ErrorCode13(suppress = self.suppress) from e


    def getArgsN(self):        
        if self.args.n:
            self.normalize = True
        else:
            self.normalize = False


    def getArgsI(self):
        try:
            if self.args.i:
                data_dir = os.path.join(self.current_wd, self.args.i)
                self.data_file = GetFiles(input_dir = data_dir)
            else:
                raise Error(code = '14')
        except Error as e:
            raise ErrorCode14(suppress = self.suppress) from e


    def getArgsUTL(self):
        args_u = [self.args.u if self.args.u else '0,0,0']
        args_l = [self.args.l if self.args.l else '0,0,0']
        args_t = [self.args.t if self.args.t else '0,0,0']

        three_tuples = [evaluateInputTuple(element) for element in tuple([args_u[0], args_l[0], args_t[0]])]
        self.tuple_u = three_tuples[0]
        self.tuple_l = three_tuples[1]
        self.tuple_t = three_tuples[2]

        ## should have at least one of the -u, -l, -t setting
        try:
            if sum(flattern(three_tuples)) == 0.0:
                raise Error(code = '18')
        except Error as e:
            raise ErrorCode18(suppress = self.suppress) from e


    def getArgsC(self):
        if self.args.c:
            try:
                if os.cpu_count() < self.args.c:  # TODO unittest
                    raise Error(code = '47')
                else:
                    self.num_cpu = self.args.c
            except Error as e:
                raise ErrorCode47(suppress = self.suppress) from e
        else:
            self.num_cpu = 1
        print(f'CPU: {self.num_cpu}\n')


    def getRevisitThresholdArgs(self):
        self.getArgsE()
        self.getArgsN()
        self.getArgsI()
        self.getArgsUTL()
        self.getArgsC()


class FindMinFromLM:
    """
    Purpose:
        Find the minimal distance between observation (y) and prediction (y_hat). The
        corresponding threshold setting (stored in input_df.index) represent the best
        option.

    Arguments:
        input_df -- Type: pandas.DataFrame.
                    Row represents sample. This dataframe contains five columns.
                    ['u'] u setting
                    ['l'] l setting
                    ['t'] t setting
                    ['x'] X axis: number of information-rich feature.
                    ['y'] Y axis:
                    (procurstes score between full_data and info_rich_subset) / (procurstes score between full_data and non_info_rich_subset)

    Attributes:
        input_df -- Type: pandas.DataFrame
                          x -- Type: numpy.array
                          y -- Type: numpy.array

                          Generated by FindMinFromLM class
                          y_hat -- Type: float;
                                   Explanation: fn(x)
                          y_hat_to_y -- Type: float;
                                        Explanation: fn(x) - y
        select_index -- Type: tuple;
                        Explanation: Row names of input_df. Store the threshold setting that generate the
                        corresponding x and x
        a -- Type: numpy.array;
             Explanation: regression coefficient.
        b -- Type: numpy.array; Explanation: intercept.
    """

    def __init__(self, input_df):
        self.input_df = input_df
        x = numpy.array(input_df['x']).reshape(len(input_df['x']), 1)
        y = numpy.array(input_df['y']).reshape(len(input_df['y']), 1)

        ## linear regression model
        lm = linear_model.LinearRegression()
        model = lm.fit(x, y)
        self.a = [0 if lm.coef_ < 10 ** (-8) else lm.coef_[0][0]][0]
        self.b = [0 if lm.intercept_ < 10 ** (-8) else lm.intercept_[0]][0]

        ## calculate distance along y axis between observation (y) and prediction (y_hat)
        self.input_df['y_hat'] = self.a * self.input_df['x'] - self.b
        self.input_df['y_hat_to_y'] = self.input_df['y_hat'] - self.input_df['y']
        select = self.input_df.loc[self.input_df['y_hat_to_y'] == max(self.input_df['y_hat_to_y'])]

        ## break tie
        self.select_index = select.index.values[0]  ## result may vary if use multicore


class RevisitThreshold:
    """
    Reselect information-rich features by using the detail_vNE files.
    """
    #def __init__(self, GetFiles_class_v, GetFiles_class_i, tuple_t, tuple_u, tuple_l, output_file_name, num_cpu, suppress = False):
    def __init__(self, GetFiles_class_v, GetFiles_class_i, tuple_t, tuple_u, tuple_l, output_file_name, num_cpu, normalize, suppress = False):
        self.detail_vNE_files = tuple(GetFiles_class_v.input_files)
        self.detail_vNE_basename = tuple([os.path.basename(element) for element in self.detail_vNE_files])
        self.detail_vNE_group = tuple([element.split("__")[0] for element in self.detail_vNE_basename])
        self.detail_vNE_group_set = tuple(sorted(list(set(self.detail_vNE_group))))

        self.data_files = tuple(GetFiles_class_i.input_files)
        self.data_file_basename = tuple([os.path.basename(element) for element in self.data_files])
        self.data_file_group = tuple(sorted(list([element.split("__")[0] for element in self.data_file_basename])))

        self.normalize = normalize
        self.suppress = suppress

        try:
            if list(self.detail_vNE_group_set) != sorted(list(set(self.data_file_group))):
                raise Error(code = '15')
        except Error as e:
            raise ErrorCode15(suppress = self.suppress) from e

        self.tuple_t = tuple_t
        self.tuple_u = tuple_u
        self.tuple_l = tuple_l
        self.output_file_name = output_file_name
        self.num_cpu = num_cpu


    #def singleFile(self, current_vNE_group, current_vNE_group_set_number, current_u_level, current_l_level):
    def singleFile(self, current_vNE_group, current_vNE_group_set_number, current_u_level, current_l_level):
        self.current_vNE = RetrospectDataImport(file_name = self.current_vNE_group[current_vNE_group_set_number], type = 'vNE', dimension = 'n')
        current_calling = InfoRichCalling(data = '', current_feature_names = '', upper_threshold_factor = current_u_level,
                                          lower_threshold_factor = current_l_level, direct_from_result_summary = self.current_vNE.vNE_summary,
                                          #num_cpu = 1, silence = True)  ## TODO: allow multiprocessing; add args
                                          num_cpu = 1, silence = True, normalize = self.normalize)
        current_calling.infoRichSelect()
        return(list(current_calling.info_rich_feature['feature_name']))


    def singleGroup(self, current_vNE_group_set_number, current_u_level, current_l_level, current_t_level):
        current_vNE_group = []
        target = self.detail_vNE_group_set[current_vNE_group_set_number]
        for i in range(len(self.detail_vNE_group)):
            if self.detail_vNE_group[i] == target:
                current_vNE_group.append(self.detail_vNE_files[i])
        self.current_vNE_group = current_vNE_group # pass x on to self.x for unittest; use x in computation ; TODO: update unittest code

        ## summarizing information-rich features identified in each iteration (each file) in to a dictionary
        revisit_infoRich_dict = {}
        for f in range(len(current_vNE_group)):
            current_info_rich_list = self.singleFile(current_vNE_group = current_vNE_group, current_vNE_group_set_number = f,
                                                     current_u_level = current_u_level, current_l_level = current_l_level)

            for element in current_info_rich_list:
                if element in revisit_infoRich_dict.keys():
                    revisit_infoRich_dict[element] += 1
                else:
                    revisit_infoRich_dict[element] = 1

        ## filter based on agrs.t (number of time that a feature need to be nominated as information-rich feature before it can be included
        ## in the final list of information-rich feature)
        revisit_infoRich_dict_filtered = {key:val for key, val in revisit_infoRich_dict.items() if val >= current_t_level}

        current_file = self.data_files[current_vNE_group_set_number]
        current_data_file = RawDataImport(file_name = current_file, for_merging_file = True, suppress = False, second_chance = False)
        current_data_file.readCSV()

        infoRich_reproducibility_at_specific_threshold_levels = reproducibility_summary(current_data_file.data, revisit_infoRich_dict_filtered)
        current_info_rich = list(infoRich_reproducibility_at_specific_threshold_levels['feature_name'])
        self.current_info_rich = current_info_rich  # pass x on to self.x for unittest; use x in computation; TODO: update unittest code

        singleGroup_return_dict = {'current_info_rich': current_info_rich, 'current_data_file_data': current_data_file.data, 'infoRich_reproducibility_at_specific_threshold_levels': infoRich_reproducibility_at_specific_threshold_levels}
        return(singleGroup_return_dict)


    def iteratesThroughGroupSet(self, current_u_level, current_l_level, current_t_level):
        info_rich_at_current_threshold_level_sub = []

        for g in range(len(self.detail_vNE_group_set)):
            singleGroup_return_dict = self.singleGroup(current_vNE_group_set_number = g, current_u_level = current_u_level,
                                                       current_l_level = current_l_level, current_t_level = current_t_level)
            current_info_rich = singleGroup_return_dict['current_info_rich']
            current_data_file_data = singleGroup_return_dict['current_data_file_data']
            infoRich_reproducibility_at_specific_threshold_levels = singleGroup_return_dict['infoRich_reproducibility_at_specific_threshold_levels']
            info_rich_at_current_threshold_level_sub.append(current_info_rich)

            if g == 0:
                merged_data = current_data_file_data   ##  TODO self.merged_data
                self.merged_data = current_data_file_data # pass x on to self.x for unittest; use x in computation; TODO: update unittest code
            else:
                merged_data = pandas.concat([merged_data, current_data_file_data], axis = 0, sort = True)  ##  TODO self.merged_data
                self.merged_data = merged_data # pass x on to self.x for unittest; use x in computation

        self.merged_data = self.merged_data.fillna(0) ##  TODO self.merged_data
        merged_data = merged_data.fillna(0) ##  TODO self.merged_data
        self.merged_data = merged_data # pass x on to self.x for unittest; use x in computation
        info_rich_at_current_threshold_level = list(set(flattern(info_rich_at_current_threshold_level_sub)))
        self.info_rich_at_current_threshold_level = info_rich_at_current_threshold_level # pass x on to self.x for unittest; use x in computation; TODO: update unittest code

        iteratesThroughGroupSet_return_dict = {'info_rich_at_current_threshold_level': info_rich_at_current_threshold_level, 'merged_data': merged_data}
        return(iteratesThroughGroupSet_return_dict)


    def compareBeforeAndAfterDataReduction(self, current_row):
        ## get current settings
        current_u_level = self.threshold_setting_summary['u'].iloc[current_row]
        current_l_level = self.threshold_setting_summary['l'].iloc[current_row]
        current_t_level = self.threshold_setting_summary['t'].iloc[current_row]

        ## iteration
        iteratesThroughGroupSet_return_dict = self.iteratesThroughGroupSet(current_u_level = current_u_level, current_l_level = current_l_level, current_t_level = current_t_level)
        info_rich_at_current_threshold_level = iteratesThroughGroupSet_return_dict['info_rich_at_current_threshold_level']
        merged_data = iteratesThroughGroupSet_return_dict['merged_data']

        ## when there are information-rich feature(s)
        if len(info_rich_at_current_threshold_level) > 0:
            # subset data
            non_info = [value for value in merged_data.columns.values if value not in info_rich_at_current_threshold_level]
            current_non_info = merged_data[non_info]
            self.current_non_info = current_non_info # pass x on to self.x for unittest; use x in computation; TODO: update unittest code

            current_info_rich = merged_data[info_rich_at_current_threshold_level]
            self.current_info_rich = current_info_rich # pass x on to self.x for unittest; use x in computation; TODO: update unittest code

            # project samples onto PCA space
            current_data_info_projection = Projection(merged_dataframe = merged_data, normalize = self.normalize).projection_df
            current_non_info_projection = Projection(merged_dataframe = current_non_info, normalize = self.normalize).projection_df
            current_info_rich_projection = Projection(merged_dataframe = current_info_rich, normalize = self.normalize).projection_df

            # procrustes test
            if current_info_rich_projection.shape[1] > 1:
                if current_info_rich_projection.shape[1] > 2:
                     d = 3
                else:
                     d = 2

                data_projection = numpy.array(current_data_info_projection.iloc[:, 0:d])
                non_info_projection = numpy.array(current_non_info_projection.iloc[:, 0:d])
                info_projection = numpy.array(current_info_rich_projection.iloc[:, 0:d])

                current_non_info_projection_disparity = procrustes(data_projection, non_info_projection)[2]
                current_info_projection_disparity = procrustes(data_projection, info_projection)[2]

            else:
                current_non_info_projection_disparity = numpy.nan
                current_info_projection_disparity = numpy.nan

        ## when there is no information-rich feature
        else:
            current_non_info_projection_disparity = 0
            current_info_projection_disparity = numpy.nan

        # ['current_u_level', 'current_l_level', 'current_t_level', 'num_info_rich', 'current_info_projection_disparity', 'current_non_info_projection_disparity']
        compareBeforeAndAfterDataReduction_return_list = [current_u_level, current_l_level, current_t_level, info_rich_at_current_threshold_level, current_info_projection_disparity, current_non_info_projection_disparity]
        return(compareBeforeAndAfterDataReduction_return_list)


    def iteratesThroughThresholdSetting(self):
        self.u_list = floatRange(self.tuple_u)
        self.l_list = floatRange(self.tuple_l)
        self.t_list = floatRange(self.tuple_t)

        self.iter_num = len(self.u_list) * len(self.l_list) * len(self.t_list)

        u_l = list(itertools.product(self.u_list, self.l_list))
        u_l_t_nested = list(itertools.product(self.t_list, u_l))
        threshold_setting_nested = pandas.DataFrame(u_l_t_nested, columns=['t', 'u_l'])
        threshold_setting_nested['u'] = threshold_setting_nested['u_l'].str[0]
        threshold_setting_nested['l'] = threshold_setting_nested['u_l'].str[1]
        threshold_setting = threshold_setting_nested
        threshold_setting = threshold_setting.drop(columns = ['u_l'])

        summary = pandas.DataFrame(data = emptyNumpyArray(nrow = self.iter_num, ncol = 3),
                                   columns = ['num_info_rich', 'info_to_ori_disparity', 'non_info_to_ori_disparity'])
        self.threshold_setting_summary = pandas.concat([threshold_setting, summary], axis = 1, sort = True)

        nrow = self.threshold_setting_summary.shape[0]
        threshold_setting_summary_result = [[] for i in range(nrow)]

        print('Precent threshold condition tested:')

        with Pool(processes = self.num_cpu) as p:
            with tqdm(total = nrow) as pbar:
                for i, res in enumerate(p.imap_unordered(self.compareBeforeAndAfterDataReduction, range(nrow))):
                    res[3] = len(list(set(res[3])))
                    threshold_setting_summary_result[i] = res
                    pbar.update()

        threshold_setting_result = pandas.DataFrame(data = threshold_setting_summary_result,
                                                     columns = ['u', 'l', 't', 'num_info_rich',
                                                                'info_to_ori_disparity', 'non_info_to_ori_disparity'])

        threshold_setting_result.to_csv(self.output_file_name)
        self.threshold_setting_result = threshold_setting_result
        return(threshold_setting_result)

    def compareSettings(self):
        self.threshold_setting_result['y'] = self.threshold_setting_result['info_to_ori_disparity'] / self.threshold_setting_result['non_info_to_ori_disparity']
        x_y_df = self.threshold_setting_result.loc[:,['u', 'l', 't', 'num_info_rich','y']]
        self.x_y_df = x_y_df.rename(columns = {'num_info_rich': 'x'})

        self.x_y_df = self.x_y_df.fillna(0)
        x_y_df_no_zero_in_x = self.x_y_df[self.x_y_df['x'] != 0]

        try:
            if x_y_df_no_zero_in_x.shape[0] >= 2:
                select_index = FindMinFromLM(x_y_df_no_zero_in_x)
                self.selected = x_y_df_no_zero_in_x.iloc[select_index.select_index]
            elif x_y_df_no_zero_in_x.shape[0] == 1:
                self.selected = x_y_df_no_zero_in_x
            else:
                raise Error(code = '27')
        except Error as e:
            raise ErrorCode27(suppress = self.suppress) from e


def revisitThresholdResult(args, current_wd, retrospect_dir, output_file_tag, suppress, silence):
    ## take-in args
    revisit_threshold_args = RevisitThresholdArgs(args = args, current_wd = current_wd, suppress = suppress, silence = silence)
    revisit_threshold_args.getRevisitThresholdArgs()

    print(args.o)

    output_file_name = os.path.join(retrospect_dir, (output_file_tag + '_threshold_setting_summary.csv'))

    ## revisit threshold and information-rich feature calling
    revisit_threshold = RevisitThreshold(GetFiles_class_v = revisit_threshold_args.detail_vNE, GetFiles_class_i = revisit_threshold_args.data_file,
                                         tuple_t = revisit_threshold_args.tuple_t, tuple_u = revisit_threshold_args.tuple_u,
                                         tuple_l = revisit_threshold_args.tuple_l, num_cpu = revisit_threshold_args.num_cpu,
                                         normalize = revisit_threshold_args.normalize, output_file_name = output_file_name)
    revisit_threshold.iteratesThroughThresholdSetting()
    revisit_threshold.compareSettings()
    print(f'\nSuggested threshold setting: -u {revisit_threshold.selected["u"]} -l {revisit_threshold.selected["l"]} -t {revisit_threshold.selected["t"]}')



#!/usr/bin/env python3

## usage
# at a level above emmer/
# python3 -m emmer.test.test_reproducibility

from ..bake import BakeCommonArgs
from ..main.basic.read import RetrospectDataImport, GetFiles
from ..posthoc.stats.reproducibility import ReproducibilitySummary, ReproducibilityArgs
from ..troubleshoot.err.error import *
from ..troubleshoot.warn.warning import *

import unittest
import numpy
import sys
import os

class TestReproducibilityArgs(unittest.TestCase):

    def test_ReproducibilityArgs(self):
        print('\ntest_ReproducibilityArgs:')
        print('        case 1: missing args.b setting')
        sys.argv[1:] = ['-m', 'Reproducibility', '-i', 'emmer/data/bake_data_dir_4/information_rich_features_summary.csv']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()
        reproducibility_args = ReproducibilityArgs(args = processed_args.args, current_wd = current_wd,
        	                                       suppress = True, silence = False)
        my_result = reproducibility_args.warning_code
        expected_result = '7'
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('        case 3: missing args.i setting')
        sys.argv[1:] = ['-m', 'Reproducibility', '-b', '40']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()
        with self.assertRaises(ErrorCode1):
            reproducibility_args = ReproducibilityArgs(args = processed_args.args, current_wd = current_wd,
                                                       suppress = True, silence = False)
            reproducibility_args.getArgsI()


        print('        ---------------------------------------------------')
        print('        case 4: expect to work')
        print('             4.1: single file')
        sys.argv[1:] = ['-m', 'Reproducibility', '-b', '40', '-i', 'emmer/data/bake_data_dir_4/information_rich_features_summary.csv']
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        current_wd = os.getcwd()
        processed_args.getHomeKeepingArgs()
        reproducibility_args = ReproducibilityArgs(args = processed_args.args, current_wd = current_wd,
                                                   suppress = True, silence = False)

        my_result = numpy.mean(reproducibility_args.reproducibility_list)
        expected_result = 88.4877777777778
        self.assertAlmostEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             4.1.1: handle bin_num')
        my_result = reproducibility_args.bin_num
        expected_result = 40
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             4.2: multiple files')
        print('             4.2.1: summarized reproducibilty output')
        sys.argv[1:] = ['-m', 'Reproducibility', '-b', '20', '-i', 'emmer/data/bake_data_dir_5']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()
        reproducibility_args = ReproducibilityArgs(args = processed_args.args, current_wd = current_wd,
                                                       suppress = True, silence = False)
        my_result = numpy.mean(reproducibility_args.reproducibility_list)
        expected_result = 88.7938888888889
        self.assertAlmostEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             4.2.2: reproducibilty output from a single input file')
        sys.argv[1:] = ['-m', 'Reproducibility', '-b', '20', '-i', 'emmer/data/bake_data_dir_8/case2_in_folder/']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()
        reproducibility_args = ReproducibilityArgs(args = processed_args.args, current_wd = current_wd,
                                                       suppress = True, silence = False)

        my_result = numpy.mean(reproducibility_args.reproducibility_list)
        expected_result = 79.1666666671568
        self.assertAlmostEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             4.3: generate a summarizing dataframe if there are multiple input file')
        print('                  row: information-rich features; column: input file names')
        sys.argv[1:] = ['-m', 'Reproducibility', '-b', '20', '-i', 'emmer/data/bake_data_dir_8/case3_in_folder/']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()
        reproducibility_args = ReproducibilityArgs(args = processed_args.args, current_wd = current_wd,
                                                   suppress = True, silence = False)
        my_result = reproducibility_args.reproducibility_summary_df.shape[0]
        expected_result = 9
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


class TestReproducibility(unittest.TestCase):

    def test_FatternInput(self):
        print('\ntest_Reproducibility.FatternInput:')
        print('    -i: "emmer/data/retrospect_data_dir_1/information_rich_features_summary.csv"')
        input_dir = 'emmer/data/retrospect_data_dir_1/information_rich_features_summary.csv'
        print('    -s: "reproducibility"')
        style = 'reproducibility'

        current_wd = os.getcwd()
        repro = RetrospectDataImport(file_name = os.path.join(current_wd, input_dir), type = style)
        repro_summary = ReproducibilitySummary(repro.reproducibility)

        my_result = list(repro_summary.reproducibility[0:6])

        expected_result = [100.0, 94.12, 0.0, 11.76, 0.0, 100.0]
        self.assertListEqual(my_result, expected_result)
        print('===========================================================')


    def test_RemoveZero(self):
        print('\ntest_Reproducibility.RemoveZero:')
        print('        case 1: post-removal result')
        print('             1.1: check the value of elements that are keep after the removal')
        print('    -i: "emmer/data/retrospect_data_dir_1/information_rich_features_summary.csv"')
        input_dir = 'emmer/data/retrospect_data_dir_1/information_rich_features_summary.csv'
        print('    -s: "reproducibility"')
        style = 'reproducibility'

        current_wd = os.getcwd()
        repro = RetrospectDataImport(file_name = os.path.join(current_wd, input_dir), type = style)
        repro_summary = ReproducibilitySummary(repro.reproducibility)

        # test 1
        my_result = list(repro_summary.reproducibility_no_zero[0:6])
        expected_result = [100.0, 94.12, 11.76, 100.0, 84.62, 100.0]
        self.assertListEqual(my_result, expected_result)


        # test 2
        print('        ---------------------------------------------------')
        print('             1.2: check the number of elements that are keep after the removal')
        my_result = len(repro_summary.reproducibility_no_zero)
        expected_result = 20
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


    def test_Stat(self):
        print('\ntest_Reproducibility.Stat:')
        print('        case 1: result')
        print('    -i: "emmer/data/retrospect_data_dir_1/information_rich_features_summary.csv"')
        input_dir = 'emmer/data/retrospect_data_dir_1/information_rich_features_summary.csv'
        print('    -s: "reproducibility"')
        style = 'reproducibility'

        current_wd = os.getcwd()
        repro = RetrospectDataImport(file_name = os.path.join(current_wd, input_dir), type = style)
        repro_summary = ReproducibilitySummary(repro.reproducibility)

        # test 1
        print('             1.1: median')
        my_result = numpy.median(repro_summary.reproducibility_no_zero)
        expected_result = 84.62
        numpy.testing.assert_almost_equal(my_result, expected_result)


        # test 2
        print('        ---------------------------------------------------')
        print('             1.2: mean')
        my_result = numpy.mean(repro_summary.reproducibility_no_zero)
        expected_result = 75.95150000000001
        numpy.testing.assert_almost_equal(my_result, expected_result)


        # test 3
        print('        ---------------------------------------------------')
        print('             1.3: standard deviation')
        my_result = numpy.std(repro_summary.reproducibility_no_zero)
        expected_result = 28.751601916241118
        numpy.testing.assert_almost_equal(my_result, expected_result)
        print('===========================================================')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

## usage
# at a level above emmer/
# python3 -m emmer.test.test_permanova

from ..bake import BakeCommonArgs
from ..main.basic.read import RetrospectDataImport, GetFiles
from ..posthoc.stats.permanova import PermanovaArgs, permanovaResult
from ..troubleshoot.err.error import *
#from ..troubleshoot.warn.warning import *

import unittest
import numpy
import sys
import os

class TestPermanovaArgs(unittest.TestCase):

    def test_PermanovaArgs(self):
        print('\ntest_PermanovaArgs:')
        print('        case 1: error and warning handling')
        print('             1.1: give both args.i and args.p')
        sys.argv[1:] = ['-m', 'Permanova', '-i', 'emmer/data/bake_data_dir_6/filtered_infoRich__PCA_coordinates.csv',
                        '-p', 'emmer/data/bake_data_dir_6/_retrospect_permanova_parameter.csv']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        permanova_args = PermanovaArgs(args = processed_args.args, current_wd = current_wd,
         	                           suppress = True, silence = False)
        my_result = permanova_args.warning_code
        expected_result = '13'
        self.assertEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             1.2: no args.i nor args.p')
        sys.argv[1:] = ['-m', 'Permanova']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode8):
            permanova_args = PermanovaArgs(args = processed_args.args, current_wd = current_wd,
            	                           suppress = True, silence = False)

        print('        ---------------------------------------------------')
        print('        case 2: expect to work')
        sys.argv[1:] = ['-m', 'Permanova', '-p', 'emmer/data/bake_data_dir_6/_retrospect_permanova_parameter.csv']
        current_wd = os.getcwd()
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        permanova_args = PermanovaArgs(args = processed_args.args, current_wd = current_wd,
            	                       suppress = True, silence = False)

        my_restul = len(permanova_args.cluster)
        expected_result = 30

        self.assertEqual(my_restul, expected_result)
        print('===========================================================')


if __name__ == '__main__':
    unittest.main()



## usage
# at a level above emmer/
# python3 -m emmer.test.test_revisit_threshold

from ..main.basic.read import RawDataImport, RetrospectDataImport, GetFiles
from ..main.advanced.iteration import MinusOneVNE, InfoRichCalling, reproducibility_summary
from ..bake import BakeCommonArgs
from ..posthoc.stats.revisit_thresholds import RevisitThresholdArgs, FindMinFromLM, evaluateInputTuple, floatRange, RevisitThreshold
from ..troubleshoot.err.error import *
#from ..troubleshoot.warn.warning import *

from itertools import compress
import numpy.testing
import argparse
import unittest
import pandas
import numpy
import time
import sys
import os



class TestRevisitThresholdArgs(unittest.TestCase):

    def test_getArgsV(self):
        print('\ntest_RevisitThreshold.getArgsN:')
        print('        case 1: missing args.n setting')
        sys.argv[1:] = ['-m', 'RevisitThreshold']
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode13):
            revisit_threshold_args = RevisitThresholdArgs(args = processed_args.args, current_wd = '', suppress = True, silence = False)
            revisit_threshold_args.getArgsE()
        print('===========================================================')


    def test_getArgsI(self):
        print('\ntest_RevisitThreshold.getArgsI:')
        print('        case 1: missing args.i setting')
        sys.argv[1:] = ['-m', 'RevisitThreshold']
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode14):
            revisit_threshold_args = RevisitThresholdArgs(args = processed_args.args, current_wd = '', suppress = True, silence = False)
            revisit_threshold_args.getArgsI()
        print('===========================================================')


    def test_getArgsUTL(self):
        print('\ntest_RevisitThreshold.getArgsUTL:')
        print('        case 1: have no args.l, args.u, and args.t setting')
        sys.argv[1:] = ['-m', 'RevisitThreshold']
        processed_args = BakeCommonArgs(suppress = True, test = False, neglect = True, silence = False)
        processed_args.getHomeKeepingArgs()

        with self.assertRaises(ErrorCode18):
            revisit_threshold_args = RevisitThresholdArgs(args = processed_args.args, current_wd = '', suppress = True, silence = False)
            revisit_threshold_args.getArgsUTL()
        print('===========================================================')


    def test_getRevisitThresholdArgs(self):   # TODO
        pass


class TestEvaluateInputTuple(unittest.TestCase):

    def test_passing(self):
        print('\ntest_EvaluateInputTuple.passing:')
        print('        case 1: passing test condition')
        print('             1.1: an reasonalbe input that != 0,0,0')
        print('        -t: "2,1,1"')
        args_t = '2,1,0.5'

        my_result = evaluateInputTuple(args_t)
        expected_result = (2,1,0.5)
        self.assertEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             1.2: an input that == 0,0,0')
        print('        -t: "0,0,0"')
        args_t = '0,0,0'

        my_result = evaluateInputTuple(args_t)
        expected_result = (0,0,0)
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


    def test_error17(self):
        print('\ntest_EvaluateInputTuple.error17:')
        print('        -t: "2,2,2,1"')
        args_t = '2,2,2,1'

        with self.assertRaises(ErrorCode17):
            output = evaluateInputTuple(args_t, suppress = True, second_chance = False)
        print('===========================================================')


    def test_error20(self):
        print('\ntest_EvaluateInputTuple.error20:')
        print('        Case 1: first element less than 0')
        print('        -t: "-2,2,1"')
        args_t = '-2,2,1'

        with self.assertRaises(ErrorCode20):
            output = evaluateInputTuple(args_t, suppress = True, second_chance = False)


        ## Case 1: confirm the function generate a empty array
        print('        ---------------------------------------------------')
        print('        Case 2: second element equals to 0')
        print('        -t: "2,0,1"')
        args_t = '2,0,1'

        with self.assertRaises(ErrorCode20):
            output = evaluateInputTuple(args_t, suppress = True, second_chance = False)
        print('===========================================================')


    def test_error19(self):
        print('\ntest_EvaluateInputTuple.error19:')
        print('        -t: "1,2,1"')
        args_t = '1,2,1'

        with self.assertRaises(ErrorCode19):
            output = evaluateInputTuple(args_t, suppress = True, second_chance = False)
        print('===========================================================')


    def test_error24(self):
        print('\ntest_EvaluateInputTuple.error24:')
        print('        -t: "3,2,0"')
        args_t = '3,2,0'

        with self.assertRaises(ErrorCode24):
            output = evaluateInputTuple(args_t, suppress = True, second_chance = False)
        print('===========================================================')


    def test_error22(self):
        print('\ntest_EvaluateInputTuple.error22:')
        print('        -t: "3,2,2"')
        args_t = '3,2,2'

        with self.assertRaises(ErrorCode22):
            output = evaluateInputTuple(args_t, suppress = True, second_chance = False)
        print('===========================================================')


    def test_error25(self):
        print('\ntest_EvaluateInputTuple.error25:')
        print('        -t: "3,2,7"')
        args_t = '3,2,7'

        with self.assertRaises(ErrorCode25):
            output = evaluateInputTuple(args_t, suppress = True, second_chance = False)
        print('===========================================================')


    def test_error26(self):
        print('\ntest_EvaluateInputTuple.error26:')
        print('        -t: "2,2,7"')
        args_t = '2,2,7'

        with self.assertRaises(ErrorCode26):
            output = evaluateInputTuple(args_t, suppress = True, second_chance = False)
        print('===========================================================')


class TestFloatRange(unittest.TestCase):

    def test_floatRange(self):
        print('\ntest_floatRange:')
        print('        input_tuple = (3, 1, 0.5)')
        input_tuple = (3, 1, 0.5)

        my_result = floatRange(input_tuple)
        expected_result = [1, 1.5, 2, 2.5, 3]
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


class TestFindMinFromLM(unittest.TestCase):

    def test_FindMinFromLM(self):
        print('\ntest_FindMinFromLM:')
        data = {'x':[1, 2, 3, 4, 5, 6], 'y':[2.2, 4.2, 6.2, 8.2, 10.2, 12.2]}
        input_df = pandas.DataFrame(data, index =['sample1', 'sample2', 'sample3', 'sample4', 'sample5', 'sample6'])

        on_original_axises = FindMinFromLM(input_df)

        print('        Case 1: linear regression')
        print('             1.1: regression correlation')
        my_result = on_original_axises.a
        expected_result = 2
        numpy.testing.assert_almost_equal(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             1.2: intercept')
        my_result = on_original_axises.b
        expected_result = 0.2
        numpy.testing.assert_almost_equal(my_result, expected_result)

        print('')
        print('        Case 2: find entry that has the minimal distance from regression line')
        print('        ---------------------------------------------------')
        print('             2.1: only one minimal value')
        data = {'x':[1, 2, 3, 4, 5, 6], 'y':[2, 3.9, 6.5, 8, 10.7, 11.5]}
        input_df = pandas.DataFrame(data, index =['sample1', 'sample2', 'sample3', 'sample4', 'sample5', 'sample6'])
        on_original_axises = FindMinFromLM(input_df)

        my_result = list([on_original_axises.select_index])
        expected_result = ['sample6']
        self.assertEqual(my_result, expected_result)


        print('        ---------------------------------------------------')
        print('             2.2: denominator')
        data = {'x':[1, 2, 3, 4, 5, 6, 6], 'y':[2, 3.9, 6.5, 8, 10.7, 11.5, 11.5]}
        input_df = pandas.DataFrame(data, index =['sample1', 'sample2', 'sample3', 'sample4', 'sample5', 'sample6', 'sample7'])
        on_original_axises = FindMinFromLM(input_df)

        my_result = list([on_original_axises.select_index])
        expected_result = ['sample6']
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


class TestRevisitThreshold(unittest.TestCase):

    def test_getFileV(self):
        print('\ntest_RevisitThreshold.getFileV:')
        print('        -v: "emmer/data/retrospect_data_dir_2/detail_vNE/"]')
        args_v = 'emmer/data/retrospect_data_dir_2/detail_vNE/'
        detail_vNE = GetFiles(input_dir = args_v)
        detail_vNE_basename = list([os.path.basename(element) for element in detail_vNE.input_files])
        detail_vNE_group = list([element.split("__")[0] for element in detail_vNE_basename])

        my_result = len(detail_vNE_group)
        expected_result = 30

        self.assertEqual(my_result, expected_result)
        print('===========================================================')


    def test_getFileI(self):
        print('\ntest_RevisitThreshold.getFileI:')
        print('        -i: "emmer/data/retrospect_data_dir_2/pre_filter_data/"')
        args_i = 'emmer/data/retrospect_data_dir_2/pre_filter_data/'
        data_dir = os.path.join(os.getcwd(), args_i)
        data_file = GetFiles(input_dir = data_dir)
        data_file_basename = list([os.path.basename(element) for element in data_file.input_files])
        data_file_group = list([element.split("__")[0] for element in data_file_basename])

        my_result = len(data_file_group)
        expected_result = 2

        self.assertEqual(my_result, expected_result)
        print('===========================================================')


    def test_error15(self):
        print('\ntest_RevisitThreshold.init:')
        print('        case 1: check input -v and -i')
        print('        -v: "emmer/data/retrospect_data_dir_2/detail_vNE/"]')
        args_v = 'emmer/data/retrospect_data_dir_2/detail_vNE/'
        input_dir = os.path.join(os.getcwd(), args_v)
        detail_vNE = GetFiles(input_dir = args_v)
        print('        -i: "emmer/data/retrospect_data_dir_2/pre_filter_data_not_match/"')
        args_i = 'emmer/data/retrospect_data_dir_2/pre_filter_data_cause_error_15/'
        data_dir = os.path.join(os.getcwd(), args_i)
        data_file = GetFiles(input_dir = data_dir)
        print('        -t: "2,1,1"')
        args_t = '2,1,1'
        tuple_t = evaluateInputTuple(args_t)
        print('        -u: "2,1,1"')
        args_u = '2,1,1'
        tuple_u = evaluateInputTuple(args_u)
        print('        -l: "2,1,1')
        args_l = '2,1,1'
        tuple_l = evaluateInputTuple(args_l)
        print('        output_file_name: "test.csv"')
        output_file_name = 'test.csv'
        print('        current_vNE_group_set_number: 1')
        current_vNE_group_set_number = 1
        print('    normalize: False')
        normalize = False

        if os.cpu_count() > 1:
            num_cpu = os.cpu_count() - 1
            print(f'    num_cpu: {num_cpu}')
        else:
            print('    num_cpu: 1')
            num_cpu = 1
            print('    your computer only have one CPU; unable to test multiprocessing function')

        with self.assertRaises(ErrorCode15):
            revisit_threshold = RevisitThreshold(GetFiles_class_v = detail_vNE, GetFiles_class_i = data_file,
                                                 tuple_t = tuple_t, tuple_u = tuple_u, tuple_l = tuple_l,
                                                 output_file_name = output_file_name, num_cpu = num_cpu,
                                                 normalize = normalize, suppress = True)


    def test_singleFile(self):
        print('\ntest_RevisitThreshold.singleFile:')

        ## case 1: check the number of file for a specific group
        #       1.1: group B
        print('        case 1: check the number of file for a specific group')
        print('             1.1: group B')
        print('        -v: "emmer/data/retrospect_data_dir_2/detail_vNE/"]')
        args_v = 'emmer/data/retrospect_data_dir_2/detail_vNE/'
        input_dir = os.path.join(os.getcwd(), args_v)
        detail_vNE = GetFiles(input_dir = args_v)
        print('        -i: "emmer/data/retrospect_data_dir_2/pre_filter_data/"')
        args_i = 'emmer/data/retrospect_data_dir_2/pre_filter_data/'
        data_dir = os.path.join(os.getcwd(), args_i)
        data_file = GetFiles(input_dir = data_dir)
        print('        -t: "3,1,1"')
        args_t = '3,1,1'
        tuple_t = evaluateInputTuple(args_t)
        print('        -u: "3,1,1"')
        args_u = '3,1,1'
        tuple_u = evaluateInputTuple(args_u)
        print('        -l: "2,1,1')
        args_l = '2,1,1'
        tuple_l = evaluateInputTuple(args_l)
        print('        output_file_name: "test216.csv"')
        output_file_name = 'test216.csv'
        print('        current_vNE_group_set_number: 1')
        current_vNE_group_set_number = 1
        if os.cpu_count() > 1:
            num_cpu = os.cpu_count() - 1

            print(f'    num_cpu: {num_cpu}')
        else:
            print('    num_cpu: 1')
            num_cpu = 1
            print('    your computer only have one CPU; unable to test multiprocessing function')

        print('    normalize: False')
        normalize = False

        revisit_threshold = RevisitThreshold(GetFiles_class_v = detail_vNE, GetFiles_class_i = data_file,
                                             tuple_t = tuple_t, tuple_u = tuple_u, tuple_l = tuple_l,
                                             output_file_name = output_file_name, num_cpu = num_cpu,
                                             normalize = normalize)

        revisit_threshold.current_vNE_group = []
        target = revisit_threshold.detail_vNE_group_set[current_vNE_group_set_number]
        for i in range(len(revisit_threshold.detail_vNE_group)):
            if revisit_threshold.detail_vNE_group[i] == target:
                revisit_threshold.current_vNE_group.append(revisit_threshold.detail_vNE_files[i])

        my_result = len(revisit_threshold.current_vNE_group)
        expected_result = 17
        self.assertEqual(my_result, expected_result)


        ##      1.2 group A
        print('        ---------------------------------------------------')
        print('             1.2: group A')
        print('        current_vNE_group_set_number: 0')
        current_vNE_group_set_number = 0

        revisit_threshold = RevisitThreshold(GetFiles_class_v = detail_vNE, GetFiles_class_i = data_file,
                                             tuple_t = tuple_t, tuple_u = tuple_u, tuple_l = tuple_l,
                                             output_file_name = output_file_name, num_cpu = num_cpu,
                                             normalize = normalize)

        revisit_threshold.current_vNE_group = []
        target = revisit_threshold.detail_vNE_group_set[current_vNE_group_set_number]
        for i in range(len(revisit_threshold.detail_vNE_group)):
            if revisit_threshold.detail_vNE_group[i] == target:
                revisit_threshold.current_vNE_group.append(revisit_threshold.detail_vNE_files[i])

        my_result = len(revisit_threshold.current_vNE_group)
        expected_result = 13
        self.assertEqual(my_result, expected_result)


        ## case 2: select the information-rich features based on different threshold
        #       2.1 keep the parameter the same as the time I first computed the information-rich feature
        print('        ---------------------------------------------------')
        print('        case 2: select the information-rich features based on different threshold')
        print('             2.1 keep the parameter the same as the time I first computed the information-rich feature')
        my_result = revisit_threshold.singleFile(current_vNE_group = revisit_threshold.current_vNE_group,
                                                 current_vNE_group_set_number = current_vNE_group_set_number,
                                                 current_u_level = 2, current_l_level = 2)

        expected_result = ['ASV_1', 'ASV_2', 'ASV_3', 'ASV_19', 'ASV_21', 'ASV_22', 'ASV_126']
        self.assertListEqual(my_result, expected_result)

        print('        ---------------------------------------------------')
        print('             2.2 change the parameter')
        #       2.2 change the parameter
        my_result = revisit_threshold.singleFile(current_vNE_group = revisit_threshold.current_vNE_group,
                                                 current_vNE_group_set_number = current_vNE_group_set_number,
                                                 current_u_level = 3, current_l_level = 3)

        expected_result = ['ASV_1', 'ASV_2', 'ASV_3', 'ASV_21']
        self.assertListEqual(my_result, expected_result)
        print('===========================================================')


    def test_singleGroup(self):
        print('\ntest_RevisitThreshold.singleGroup:')

        ## case 1: select the information-rich features based on different threshold
        #       1.1: keep the parameter the same as the time I first computed the information-rich feature
        print('        case 1: select the information-rich features based on different threshold')
        print('             1.1: keep the parameter the same as the time I first computed the information-rich feature')

        print('        -v: "emmer/data/retrospect_data_dir_2/detail_vNE/"]')
        args_v = 'emmer/data/retrospect_data_dir_2/detail_vNE/'
        input_dir = os.path.join(os.getcwd(), args_v)
        detail_vNE = GetFiles(input_dir = args_v)

        print('        -i: "emmer/data/retrospect_data_dir_2/pre_filter_data/"')
        args_i = 'emmer/data/retrospect_data_dir_2/pre_filter_data/'
        data_dir = os.path.join(os.getcwd(), args_i)
        data_file = GetFiles(input_dir = data_dir)

        print('        -t: "3,1,1"')
        args_t = '3,1,1'
        tuple_t = evaluateInputTuple(args_t)
        print('        -u: "3,1,1"')
        args_u = '3,1,1'
        tuple_u = evaluateInputTuple(args_u)
        print('        -l: "2,1,1')
        args_l = '2,1,1'
        tuple_l = evaluateInputTuple(args_l)
        print('        current_vNE_group_set_number: 0')
        current_vNE_group_set_number = 0
        print('        current_u_level: 2')
        current_u_level = 2
        print('        current_l_level: 2')
        current_l_level = 2
        print('        current_t_level: 2')
        current_t_level = 2
        print('        output_file_name: "test310.csv"')
        output_file_name = 'test310.csv'
        if os.cpu_count() > 1:
            num_cpu = os.cpu_count() - 1

            print(f'    num_cpu: {num_cpu}')
        else:
            print('    num_cpu: 1')
            num_cpu = 1
            print('    your computer only have one CPU; unable to test multiprocessing function')
        print('    normalize: False')
        normalize = False

        revisit_threshold = RevisitThreshold(GetFiles_class_v = detail_vNE, GetFiles_class_i = data_file,
                                             tuple_t = tuple_t, tuple_u = tuple_u, tuple_l = tuple_l,
                                             output_file_name = output_file_name, num_cpu = num_cpu,
                                             normalize = normalize)

        revisit_threshold.singleGroup(current_vNE_group_set_number = current_vNE_group_set_number,
                                      current_u_level = current_u_level, current_l_level = current_l_level,
                                      current_t_level = current_t_level)

        my_result = sorted(revisit_threshold.current_info_rich)
        expected_result = ['ASV_1', 'ASV_126', 'ASV_19', 'ASV_2', 'ASV_21', 'ASV_22', 'ASV_23', 'ASV_3', 'ASV_78']
        self.assertListEqual(my_result, expected_result)

        #       1.2 change the parameter
        print('        ---------------------------------------------------')
        print('             1.2 change the parameter')
        print('        current_u_level: 3')
        current_u_level = 3
        print('        current_l_level: 3')
        current_l_level = 3
        print('        current_t_level: 2')
        current_t_level = 2
        print('        output_file_name: "test333.csv"')
        output_file_name = 'test333.csv'

        revisit_threshold = RevisitThreshold(GetFiles_class_v = detail_vNE, GetFiles_class_i = data_file,
                                             tuple_t = tuple_t, tuple_u = tuple_u, tuple_l = tuple_l,
                                             output_file_name = output_file_name, num_cpu = num_cpu,
                                             normalize = normalize)

        revisit_threshold.singleGroup(current_vNE_group_set_number = current_vNE_group_set_number,
                                      current_u_level = current_u_level, current_l_level = current_l_level,
                                      current_t_level = current_t_level)

        my_result = sorted(revisit_threshold.current_info_rich)
        expected_result = ['ASV_1', 'ASV_126', 'ASV_2', 'ASV_21', 'ASV_3', 'ASV_78']
        self.assertListEqual(my_result, expected_result)
        print('===========================================================')


    def test_iteratesThroughGroupSet(self):
        print('\ntest_RevisitThreshold.iteratesThroughGroupSet:')

        ## Case 1: combine lists of information-rich features
        print('        Case 1: combine lists of information-rich features')
        print('        -v: "emmer/data/retrospect_data_dir_2/detail_vNE/"]')
        args_v = 'emmer/data/retrospect_data_dir_2/detail_vNE/'
        input_dir = os.path.join(os.getcwd(), args_v)
        detail_vNE = GetFiles(input_dir = args_v)
        print('        -i: "emmer/data/retrospect_data_dir_2/pre_filter_data/"')
        args_i = 'emmer/data/retrospect_data_dir_2/pre_filter_data/'
        data_dir = os.path.join(os.getcwd(), args_i)
        data_file = GetFiles(input_dir = data_dir)
        print('        -t: "2,1,1"')
        args_t = '2,1,1'
        tuple_t = evaluateInputTuple(args_t)
        print('        -u: "3,1,1"')
        args_u = '3,1,1'
        tuple_u = evaluateInputTuple(args_u)
        print('        -l: "3,1,1')
        args_l = '3,1,1'
        tuple_l = evaluateInputTuple(args_l)
        print('        current_u_level: 3')
        current_u_level = 3
        print('        current_l_level: 3')
        current_l_level = 3
        print('        current_t_level: 2')
        current_t_level = 2
        print('        output_file_name: "test377.csv"')
        output_file_name = 'test377.csv'
        if os.cpu_count() > 1:
            num_cpu = os.cpu_count() - 1

            print(f'    num_cpu: {num_cpu}')
        else:
            print('    num_cpu: 1')
            num_cpu = 1
            print('    your computer only have one CPU; unable to test multiprocessing function')

        print('    normalize: False')
        normalize = False

        revisit_threshold = RevisitThreshold(GetFiles_class_v = detail_vNE, GetFiles_class_i = data_file,
                                             tuple_t = tuple_t, tuple_u = tuple_u, tuple_l = tuple_l,
                                             output_file_name = output_file_name, num_cpu = num_cpu,
                                             normalize = normalize)

        revisit_threshold.iteratesThroughGroupSet(current_u_level = current_u_level,
                                                  current_l_level = current_l_level,
                                                  current_t_level = current_t_level)

        my_result = sorted(revisit_threshold.info_rich_at_current_threshold_level)
        expected_result = ['ASV_1', 'ASV_107', 'ASV_126', 'ASV_2', 'ASV_21', 'ASV_3', 'ASV_32', 'ASV_77', 'ASV_78']
        self.assertEqual(my_result, expected_result)

        ## Case 2: merge pre_filter data
        print('        ---------------------------------------------------')
        print('        Case 2: merge pre_filter data')
        my_result = revisit_threshold.merged_data.shape[1]
        expected_result = 1670
        self.assertEqual(my_result, expected_result)
        print('===========================================================')


    def test_compareBeforeAndAfterDataReduction(self):
        print('\ntest_RevisitThreshold.compareBeforeAndAfterDataReduction:')

        ## Case 1: check the number of information-rich feature
        print('    Case 1: check the number of information-rich feature')
        print('        -v: "emmer/data/retrospect_data_dir_2/detail_vNE/"]')
        args_v = 'emmer/data/retrospect_data_dir_2/detail_vNE/'
        input_dir = os.path.join(os.getcwd(), args_v)
        detail_vNE = GetFiles(input_dir = args_v)
        print('        -i: "emmer/data/retrospect_data_dir_2/pre_filter_data/"')
        args_i = 'emmer/data/retrospect_data_dir_2/pre_filter_data/'
        data_dir = os.path.join(os.getcwd(), args_i)
        data_file = GetFiles(input_dir = data_dir)
        print('        -u: "3,3,0"')
        args_u = '3,3,0'
        tuple_u = evaluateInputTuple(args_u)
        print('        -l: "3,3,0"')
        args_l = '3,3,0'
        tuple_l = evaluateInputTuple(args_l)
        print('        -t: "2,2,0"')
        args_t = '2,2,0'
        tuple_t = evaluateInputTuple(args_t)
        print('        output_file_name: "test427.csv"')
        output_file_name = 'test427.csv'
        print('    num_cpu: 1')
        num_cpu = 1
        print('    normalize: False')
        normalize = False

        revisit_threshold = RevisitThreshold(GetFiles_class_v = detail_vNE, GetFiles_class_i = data_file,
                                             tuple_t = tuple_t, tuple_u = tuple_u, tuple_l = tuple_l,
                                             output_file_name = output_file_name, num_cpu = num_cpu,
                                             normalize = normalize)
        revisit_threshold.iteratesThroughThresholdSetting()
        output = revisit_threshold.compareBeforeAndAfterDataReduction(current_row = 0)
        my_result = sorted(output[3])
        expected_result = ['ASV_1', 'ASV_107', 'ASV_126', 'ASV_2', 'ASV_21', 'ASV_3', 'ASV_32', 'ASV_77', 'ASV_78']
        self.assertListEqual(my_result, expected_result)


        ## Case 2: check data splitting
        print('        ---------------------------------------------------')
        print('        Case 2: check data splitting')
        my_result = revisit_threshold.current_non_info.shape[1] + revisit_threshold.current_info_rich.shape[1]
        expected_result = revisit_threshold.merged_data.shape[1]
        self.assertEqual(my_result, expected_result)


        ## Case 3: check procrustes test result
        print('        ---------------------------------------------------')
        print('        Case 3: check procrustes test result')
        my_result = output[4]
        expected_result = 0.005937139500067904
        numpy.testing.assert_almost_equal(my_result, expected_result, decimal = 5)
        os.remove('test427.csv')
        print('===========================================================')


    def test_iteratesThroughThresholdSetting(self):
        print('\ntest_RevisitThreshold.iteratesThroughThresholdSetting:')

        ## Case 1: Number of iteration
        print('        Case 1: Number of iteration')
        print('        -v: "emmer/data/retrospect_data_dir_2/detail_vNE/"]')
        args_v = 'emmer/data/retrospect_data_dir_2/detail_vNE/'
        input_dir = os.path.join(os.getcwd(), args_v)
        detail_vNE = GetFiles(input_dir = args_v)
        print('        -i: "emmer/data/retrospect_data_dir_2/pre_filter_data/"')
        args_i = 'emmer/data/retrospect_data_dir_2/pre_filter_data/'
        data_dir = os.path.join(os.getcwd(), args_i)
        data_file = GetFiles(input_dir = data_dir)
        print('        -t: "2,1,1"')
        args_t = '1,1,0'
        tuple_t = evaluateInputTuple(args_t)
        print('        -u: "4,2,1"')
        args_u = '3,2,1'
        tuple_u = evaluateInputTuple(args_u)
        print('        -l: "4,2,1')
        args_l = '3,2,1'
        tuple_l = evaluateInputTuple(args_l)
        print('        output_file_name: "test481.csv"')
        output_file_name = 'test481.csv'
        if os.cpu_count() > 1:
            num_cpu = os.cpu_count() - 1
            print(f'    num_cpu: {num_cpu}')
        else:
            print('    num_cpu: 1')
            num_cpu = 1
            print('    your computer only have one CPU; unable to test multiprocessing function')
        print('    normalize: False')
        normalize = False

        start_time = time.time()

        revisit_threshold = RevisitThreshold(GetFiles_class_v = detail_vNE, GetFiles_class_i = data_file,
                                             tuple_t = tuple_t, tuple_u = tuple_u, tuple_l = tuple_l,
                                             output_file_name = output_file_name, num_cpu = num_cpu,
                                             normalize = normalize)

        revisit_threshold.iteratesThroughThresholdSetting()

        my_result = revisit_threshold.iter_num
        expected_result = 4
        self.assertEqual(my_result, expected_result)
        print('Speed test for RevisitThreshold.iteratesThroughThresholdSetting')
        print('computation time: %s seconds ' % (time.time() - start_time))
        os.remove('test481.csv')


        ## Case 2: Ability to include NaN
        print('        ---------------------------------------------------')
        print('        Case 2: Ability to include NaN')
        print('        -t: "2,1,1"')
        args_t = '2,1,1'
        tuple_t = evaluateInputTuple(args_t)
        print('        -u: "30,29,1"')
        args_u = '30,29,1'
        tuple_u = evaluateInputTuple(args_u)
        print('        -l: "30,29,1')
        args_l = '30,29,1'
        tuple_l = evaluateInputTuple(args_l)
        print('        output_file_name: "test509.csv"')
        output_file_name = 'test509.csv'
        if os.cpu_count() > 1:
            num_cpu = os.cpu_count() - 1

            print(f'    num_cpu: {num_cpu}')
        else:
            print('    num_cpu: 1')
            num_cpu = 1
            print('    your computer only have one CPU; unable to test multiprocessing function')

        print('    normalize: False')
        normalize = False

        revisit_threshold = RevisitThreshold(GetFiles_class_v = detail_vNE, GetFiles_class_i = data_file,
                                             tuple_t = tuple_t, tuple_u = tuple_u, tuple_l = tuple_l,
                                             output_file_name = output_file_name, num_cpu = num_cpu,
                                             normalize = normalize)

        revisit_threshold.iteratesThroughThresholdSetting()
        my_result = list(revisit_threshold.threshold_setting_result['info_to_ori_disparity'].isnull())

        expected_result = [True, True, True, True, True, True, True, True]
        self.assertEqual(my_result, expected_result)

        my_result = revisit_threshold.threshold_setting_result['non_info_to_ori_disparity'].iloc[0]
        expected_result = 0
        numpy.testing.assert_almost_equal(my_result, expected_result)

        os.remove('test509.csv')
        print('===========================================================')


    def test_compareSettings(self):
        print('\ntest_RevisitThreshold.compareSettings:')
        print('        Case 1: Expected input threshold setting')
        print('        -v: "emmer/data/retrospect_data_dir_2/detail_vNE/"]')
        args_v = 'emmer/data/retrospect_data_dir_2/detail_vNE/'
        input_dir = os.path.join(os.getcwd(), args_v)
        detail_vNE = GetFiles(input_dir = args_v)
        print('        -i: "emmer/data/retrospect_data_dir_2/pre_filter_data/"')
        args_i = 'emmer/data/retrospect_data_dir_2/pre_filter_data/'
        data_dir = os.path.join(os.getcwd(), args_i)
        data_file = GetFiles(input_dir = data_dir)
        print('        -t: "1,1,0"')
        args_t = '1,1,0'
        tuple_t = evaluateInputTuple(args_t)
        print('        -u: "3,2,1"')
        args_u = '3,2,1'
        tuple_u = evaluateInputTuple(args_u)
        print('        -l: "3,2,1')
        args_l = '3,2,1'
        tuple_l = evaluateInputTuple(args_l)
        print('        output_file_name: "test570.csv"')
        output_file_name = 'test570.csv'
        print('    num_cpu: 1')  # order matters in this case
        num_cpu = 1
        print('    normalize: False')
        normalize = False

        revisit_threshold = RevisitThreshold(GetFiles_class_v = detail_vNE, GetFiles_class_i = data_file,
                                             tuple_t = tuple_t, tuple_u = tuple_u, tuple_l = tuple_l,
                                             output_file_name = output_file_name, num_cpu = num_cpu,
                                             normalize = normalize)

        revisit_threshold.iteratesThroughThresholdSetting()
        revisit_threshold.compareSettings()

        my_result = revisit_threshold.selected['u']
        expected_result = 2
        numpy.testing.assert_almost_equal(my_result, expected_result)

        my_result = revisit_threshold.selected['l']
        expected_result = 2

        numpy.testing.assert_almost_equal(my_result, expected_result)

        my_result = revisit_threshold.selected['t']
        expected_result = 1
        numpy.testing.assert_almost_equal(my_result, expected_result)

        os.remove('test570.csv')


        print('        ---------------------------------------------------')
        print('        Case 2: No information-rich feature can be selected at the input threshold setting')
        print('        -t: "2,1,1"')
        args_t = '2,1,1'
        tuple_t = evaluateInputTuple(args_t)
        print('        -u: "30,29,1"')
        args_u = '30,29,1'
        tuple_u = evaluateInputTuple(args_u)
        print('        -l: "30,29,1')
        args_l = '30,29,1'
        tuple_l = evaluateInputTuple(args_l)
        print('        output_file_name: "test605.csv"')
        output_file_name = 'test605.csv'
        if os.cpu_count() > 1:
            num_cpu = os.cpu_count() - 1

            print(f'    num_cpu: {num_cpu}')
        else:
            print('    num_cpu: 1')
            num_cpu = 1
            print('    your computer only have one CPU; unable to test multiprocessing function')
        print('    normalize: False')
        normalize = False

        with self.assertRaises(ErrorCode27):
            revisit_threshold = RevisitThreshold(GetFiles_class_v = detail_vNE, GetFiles_class_i = data_file,
                                                 tuple_t = tuple_t, tuple_u = tuple_u, tuple_l = tuple_l,
                                                 output_file_name = output_file_name, num_cpu = num_cpu,
                                                 normalize = normalize, suppress = True)

            revisit_threshold.iteratesThroughThresholdSetting()
            out = revisit_threshold.compareSettings()

        os.remove('test605.csv')
        print('===========================================================')


if __name__ == '__main__':
    unittest.main()
